// Copyright (c) 2001-2006, Scalable Network Technologies, Inc.  All Rights Reserved.
//                          6701 Center Drive West
//                          Suite 520
//                          Los Angeles, CA 90045
//                          sales@scalable-networks.com
//
// This source code is licensed, not sold, and is subject to a written
// license agreement.  Among other things, no portion of this source
// code may be copied, transmitted, disclosed, displayed, distributed,
// translated, used as the basis for a derivative work, or used, in
// whole or in part, for any program or purpose other than its intended
// use in compliance with the license agreement as part of the QualNet
// software.  This source code and certain of the algorithms contained
// within it are confidential trade secrets of Scalable Network
// Technologies, Inc. and may not be used as the basis for any other
// software, hardware, product or service.

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "api.h"
#include "app_util.h"
#include "app_mcbr.h"

/*
 * NAME:        AppLayerMCbrClient.
 * PURPOSE:     Models the behaviour of MCbrClient Client on receiving the
 *              message encapsulated in msg.
 * PARAMETERS:  node - pointer to the node which received the message.
 *              msg - message received by the layer
 * RETURN:      none.
 */
void
AppLayerMCbrClient(Node *node, Message *msg)
{
    char buf[MAX_STRING_LENGTH];
    AppDataMCbrClient *clientPtr;

    ctoa(getSimTime(node), buf);

#ifdef DEBUG
    printf("MCBR Client: Node %u got a message at %s\n",
           node->nodeId, buf);
#endif /* DEBUG */

    switch (msg->eventType)
    {
        case MSG_APP_TimerExpired:
        {
            AppTimer *timer;

            timer = (AppTimer *) MESSAGE_ReturnInfo(msg);

#ifdef DEBUG
            printf("MCBR Client: Node %u at %s got timer %d\n",
                   node->nodeId, buf, timer->type);
#endif /* DEBUG */

            clientPtr = AppMCbrClientGetMCbrClient(node, timer->sourcePort);
            Int32 itemSizeRead = clientPtr->itemSize;


           if (clientPtr == NULL)
            {
                printf("MCBR Client: Node %u cannot find mcbr client\n",
                       node->nodeId);

                assert(FALSE);
            }

            switch (timer->type)
            {
                case APP_TIMER_SEND_PKT:
                {
                    char *payload;
                    MCbrData data;

                    payload = (char *)MEM_malloc(itemSizeRead);
                    memset(payload, 0, itemSizeRead);

#ifdef DEBUG
                    printf("MCBR Client: Node %u has %u items left"
                        "to send\n", node->nodeId, clientPtr->itemsToSend);
#endif /* DEBUG */

                    if (((clientPtr->itemsToSend > 1) &&
                         (getSimTime(node) + clientPtr->interval
                         < clientPtr->endTime)) ||
                        ((clientPtr->itemsToSend == 0) &&
                         (getSimTime(node) + clientPtr->interval
                          < clientPtr->endTime)) ||
                        ((clientPtr->itemsToSend > 1) &&
                         (clientPtr->endTime == 0)) ||
                        ((clientPtr->itemsToSend == 0) &&
                         (clientPtr->endTime == 0)))
                    {
                        data.type = 'd';
                    }
                    else
                    {
                        data.type = 'c';
                        clientPtr->sessionIsClosed = TRUE;
                        clientPtr->sessionFinish = getSimTime(node);
                    }

                    data.sourcePort = clientPtr->sourcePort;
                    data.txTime = getSimTime(node);
                    data.seqNo = clientPtr->seqNo++;
                    memcpy(payload, &data, sizeof(data));

#ifdef DEBUG
                    {
                        char clockStr[24];
                        ctoa(getSimTime(node), clockStr);

                        printf("MCBR Client: node %u sending data packet"
                               " at time %s to MCBR server %u\n",
                               node->nodeId,
                               clockStr,
                               clientPtr->remoteAddr);
                        printf("    size of payload is %d\n",
                               clientPtr->itemSize);
                    }
#endif /* DEBUG */

                    APP_UdpSendNewDataWithPriority(
                        node,
                        APP_MCBR_SERVER,
                        clientPtr->localAddr,
                        (short) clientPtr->sourcePort,
                        clientPtr->remoteAddr,
                        ANY_INTERFACE,
                        payload,
                        clientPtr->itemSize,
                        APP_DEFAULT_TOS,
                        0,
                        TRACE_MCBR);

                    clientPtr->numBytesSent += clientPtr->itemSize;
                    clientPtr->numPktsSent++;
                    clientPtr->sessionLastSent = getSimTime(node);

                    if (clientPtr->itemsToSend > 0)
                    {
                        clientPtr->itemsToSend--;
                    }

                    if (clientPtr->sessionIsClosed == FALSE)
                    {
                        AppMCbrClientScheduleNextPkt(node, clientPtr);
                    }

                    MEM_free(payload);
                    break;
                }

                default:
                    assert(FALSE);
            }

            break;
        }
        default:
           ctoa(getSimTime(node), buf);
           printf("MCBR Client: at time %s, node %u "
                  "received message of unknown type"
                  " %u.\n", buf, node->nodeId, msg->eventType);
           assert(FALSE);
    }

    MESSAGE_Free(node, msg);
}

/*
 * NAME:        AppMCbrClientInit.
 * PURPOSE:     Initialize a MCbrClient session.
 * PARAMETERS:  node - pointer to the node,
 *              serverAddr - address of the server,
 *              itemsToSend - number of items to send,
 *              itemSize - size of each packet,
 *              interval - interval of packet transmission rate.
 *              startTime - time until the session starts,
 *              endTime - time until the session ends,
 * RETURN:      none.
 */
void
AppMCbrClientInit(
    Node *node,
    Address clientAddr,
    Address serverAddr,
    Int32 itemsToSend,
    Int32 itemSize,
    clocktype interval,
    clocktype startTime,
    clocktype endTime)
{
    AppTimer *timer;
    AppDataMCbrClient *clientPtr;
    Message *timerMsg;
    int minSize;

    minSize = sizeof(MCbrData);

    /* Check to make sure the number of mcbr items is a correct value. */
    if (itemsToSend < 0)
    {
        printf("MCBR Client: Node %u item to sends needs to be >= 0\n",
               node->nodeId);

        exit(0);
    }

    /* Make sure that packet is big enough to carry mcbr data information. */
    if (itemSize < minSize)
    {
        printf("MCBR Client: Node %u item size needs to be >= %d.\n",
                node->nodeId, minSize);
        exit(0);
    }

    /* Make sure that packet is within max limit. */
    if (itemSize > APP_MAX_DATA_SIZE)
    {
        printf("MCBR Client: Node %u item size needs to be <= %d.\n",
                node->nodeId, APP_MAX_DATA_SIZE);
        exit(0);
    }

    /* Make sure interval is valid. */
    if (interval <= 0)
    {
        printf("MCBR Client: Node %u interval needs to be > 0.\n",
                node->nodeId);
        exit(0);
    }

    /* Make sure start time is valid. */
    if (startTime < 0)
    {
        printf("MCBR Client: Node %u start time needs to be >= 0.\n",
                node->nodeId);
        exit(0);
    }

    /* Check to make sure the end time is a correct value. */
    if (!((endTime > startTime) || (endTime == 0)))
    {
        printf("MCBR Client: Node %u end time needs to be > start time "
               "or equal to 0.\n", node->nodeId);

        exit(0);
    }

    clientPtr = AppMCbrClientNewMCbrClient(node,
                                         clientAddr,
                                         serverAddr,
                                         itemsToSend,
                                         itemSize,
                                         interval,
                                         startTime,
                                         endTime);

    if (clientPtr == NULL)
    {
        printf("MCBR Client: Node %u cannot allocate memory for"
               " new client\n", node->nodeId);

        assert(FALSE);
    }

    timerMsg = MESSAGE_Alloc(node,
                              APP_LAYER,
                              APP_MCBR_CLIENT,
                              MSG_APP_TimerExpired);

    MESSAGE_InfoAlloc(node, timerMsg, sizeof(AppTimer));

    timer = (AppTimer *)MESSAGE_ReturnInfo(timerMsg);

    timer->sourcePort = (short) clientPtr->sourcePort;
    timer->type = APP_TIMER_SEND_PKT;

#ifdef DEBUG
    {
        char clockStr[24];
        ctoa(startTime, clockStr);
        printf("MCBR Client: Node %u starting client at %s\n",
               node->nodeId, clockStr);
    }
#endif /* DEBUG */

    MESSAGE_Send(node, timerMsg, startTime);
}

/*
 * NAME:        AppMCbrClientPrintStats.
 * PURPOSE:     Prints statistics of a MCbrClient session.
 * PARAMETERS:  node - pointer to the this node.
 *              clientPtr - pointer to the mcbr client data structure.
 * RETURN:      none.
 */
void AppMCbrClientPrintStats(Node *node, AppDataMCbrClient *clientPtr) {
    clocktype throughput;

    char addrStr[MAX_STRING_LENGTH];
    char startStr[MAX_STRING_LENGTH];
    char closeStr[MAX_STRING_LENGTH];
    char sessionStatusStr[MAX_STRING_LENGTH];
    char throughputStr[MAX_STRING_LENGTH];
    char numByteSentStr[MAX_STRING_LENGTH];

    char buf[MAX_STRING_LENGTH];

    TIME_PrintClockInSecond(clientPtr->sessionStart, startStr);
    TIME_PrintClockInSecond(clientPtr->sessionLastSent, closeStr);

    if (clientPtr->sessionIsClosed == FALSE) {
        clientPtr->sessionFinish = getSimTime(node);
        sprintf(sessionStatusStr, "Not closed");
    }
    else {
        sprintf(sessionStatusStr, "Closed");
    }

    if (clientPtr->sessionFinish <= clientPtr->sessionStart) {
        throughput = 0;
    }
    else {
        throughput = (clocktype)
                     ((clientPtr->numBytesSent * SECOND) /
                     (clientPtr->sessionFinish - clientPtr->sessionStart));

        throughput = throughput * 8;
    }

    ctoa(throughput, throughputStr);
    ctoa(clientPtr->numBytesSent, numByteSentStr);


    IO_ConvertIpAddressToString(&clientPtr->remoteAddr, addrStr);

    sprintf(buf, "Server Address = %s", addrStr);
    IO_PrintStat(
        node,
        "Application",
        "MCBR Client",
        ANY_DEST,
        clientPtr->sourcePort,
        buf);

    sprintf(buf, "First Packet Sent at (s) = %s", startStr);
    IO_PrintStat(
        node,
        "Application",
        "MCBR Client",
        ANY_DEST,
        clientPtr->sourcePort,
        buf);

    sprintf(buf, "Last Packet Sent at (s) = %s", closeStr);
    IO_PrintStat(
        node,
        "Application",
        "MCBR Client",
        ANY_DEST,
        clientPtr->sourcePort,
        buf);

    sprintf(buf, "Session Status = %s", sessionStatusStr);
    IO_PrintStat(
        node,
        "Application",
        "MCBR Client",
        ANY_DEST,
        clientPtr->sourcePort,
        buf);


    sprintf(buf, "Total Bytes Sent = %s", numByteSentStr);

    IO_PrintStat(
        node,
        "Application",
        "MCBR Client",
        ANY_DEST,
        clientPtr->sourcePort,
        buf);

    sprintf(buf, "Total Packets Sent = %d", clientPtr->numPktsSent);
    IO_PrintStat(
        node,
        "Application",
        "MCBR Client",
        ANY_DEST,
        clientPtr->sourcePort,
        buf);

    sprintf(buf, "Throughput (bits/s) = %s", throughputStr);
    IO_PrintStat(
        node,
        "Application",
        "MCBR Client",
        ANY_DEST,
        clientPtr->sourcePort,
        buf);
}

/*
 * NAME:        AppMCbrClientFinalize.
 * PURPOSE:     Collect statistics of a MCbrClient session.
 * PARAMETERS:  node - pointer to the node.
 *              appInfo - pointer to the application info data structure.
 * RETURN:      none.
 */
void
AppMCbrClientFinalize(Node *node, AppInfo* appInfo)
{
    AppDataMCbrClient *clientPtr = (AppDataMCbrClient*)appInfo->appDetail;

    if (node->appData.appStats == TRUE)
    {
        AppMCbrClientPrintStats(node, clientPtr);
    }
}

/*
 * NAME:        AppMCbrClientGetMCbrClient.
 * PURPOSE:     search for a mcbr client data structure.
 * PARAMETERS:  node - pointer to the node.
 *              sourcePort - connection ID of the mcbr client.
 * RETURN:      the pointer to the mcbr client data structure,
 *              NULL if nothing found.
 */
AppDataMCbrClient *
AppMCbrClientGetMCbrClient(Node *node, int sourcePort)
{
    AppInfo *appList = node->appData.appPtr;
    AppDataMCbrClient *mcbrClient;

    for (; appList != NULL; appList = appList->appNext)
    {
        if (appList->appType == APP_MCBR_CLIENT)
        {
            mcbrClient = (AppDataMCbrClient *) appList->appDetail;

            if (mcbrClient->sourcePort == sourcePort)
            {
                return mcbrClient;
            }
        }
    }

    return NULL;
}

/*
 * NAME:        AppMCbrClientNewMCbrClient.
 * PURPOSE:     create a new mcbr client data structure, place it
 *              at the beginning of the application list.
 * PARAMETERS:  node - pointer to the node.
 *              remoteAddr - remote address.
 *              itemsToSend - number of mcbr items to send in simulation.
 *              itemSize - size of each packet.
 *              interval - time between two packets.
 *              startTime - when the node will start sending.
 * RETURN:      the pointer to the created mcbr client data structure,
 *              NULL if no data structure allocated.
 */
AppDataMCbrClient *
AppMCbrClientNewMCbrClient(Node *node,
                         Address localAddr,
                         Address remoteAddr,
                         Int32 itemsToSend,
                         Int32 itemSize,
                         clocktype interval,
                         clocktype startTime,
                         clocktype endTime)
{
    AppDataMCbrClient *mcbrClient;

    mcbrClient = (AppDataMCbrClient *)
                MEM_malloc(sizeof(AppDataMCbrClient));

    /*
     * fill in mcbr info.
     */
    mcbrClient->localAddr = localAddr;
    mcbrClient->remoteAddr = remoteAddr;
    mcbrClient->interval = interval;
    mcbrClient->sessionStart = getSimTime(node) + startTime;
    mcbrClient->sessionIsClosed = FALSE;
    mcbrClient->sessionLastSent = getSimTime(node);
    mcbrClient->sessionFinish = getSimTime(node);
    mcbrClient->endTime = endTime;
    mcbrClient->numBytesSent = 0;
    mcbrClient->numPktsSent = 0;
    mcbrClient->itemsToSend = itemsToSend;
    mcbrClient->itemSize = itemSize;
    mcbrClient->sourcePort = node->appData.nextPortNum++;
    mcbrClient->seqNo = 0;

#ifdef DEBUG
    {
        char clockStr[24];
        printf("MCBR Client: Node %u created new mcbr client structure\n",
               node->nodeId);
        printf("    localAddr = %u\n", mcbrClient->localAddr);
        printf("    remoteAddr = %u\n", mcbrClient->remoteAddr);
        ctoa(mcbrClient->interval, clockStr);
        printf("    interval = %s\n", clockStr);
        ctoa(mcbrClient->sessionStart, clockStr);
        printf("    sessionStart = %s\n", clockStr);
        printf("    numBytesSent = %u\n", mcbrClient->numBytesSent);
        printf("    numPktsSent = %u\n", mcbrClient->numPktsSent);
        printf("    itemsToSend = %u\n", mcbrClient->itemsToSend);
        printf("    itemSize = %u\n", mcbrClient->itemSize);
        printf("    sourcePort = %u\n", mcbrClient->sourcePort);
        printf("    seqNo = %u\n", mcbrClient->seqNo);
    }
#endif /* DEBUG */

    APP_RegisterNewApp(node, APP_MCBR_CLIENT, mcbrClient);

    return mcbrClient;
}

/*
 * NAME:        AppMCbrClientScheduleNextPkt.
 * PURPOSE:     schedule the next packet the client will send.  If next packet
 *              won't arrive until the finish deadline, schedule a close.
 * PARAMETERS:  node - pointer to the node,
 *              clientPtr - pointer to the mcbr client data structure.
 * RETRUN:      none.
 */
void //inline//
AppMCbrClientScheduleNextPkt(Node *node, AppDataMCbrClient *clientPtr)
{
    AppTimer *timer;
    Message *timerMsg;

    timerMsg = MESSAGE_Alloc(node,
                              APP_LAYER,
                              APP_MCBR_CLIENT,
                              MSG_APP_TimerExpired);

    MESSAGE_InfoAlloc(node, timerMsg, sizeof(AppTimer));

    timer = (AppTimer *)MESSAGE_ReturnInfo(timerMsg);
    timer->sourcePort = (short)clientPtr->sourcePort;

    timer->type = APP_TIMER_SEND_PKT;

#ifdef DEBUG
    {
        char clockStr[24];
        printf("MCBR Client: Node %u scheduling next data packet\n",
               node->nodeId);
        printf("    timer type is %d\n", timer->type);
        ctoa(clientPtr->interval, clockStr);
        printf("    interval is %s\n", clockStr);
    }
#endif /* DEBUG */

    MESSAGE_Send(node, timerMsg, clientPtr->interval);
}


/*
 * NAME:        AppLayerMCbrServer.
 * PURPOSE:     Models the behaviour of MCbrServer Server on receiving the
 *              message encapsulated in msg.
 * PARAMETERS:  node - pointer to the node which received the message.
 *              msg - message received by the layer
 * RETURN:      none.
 */
void
AppLayerMCbrServer(Node *node, Message *msg)
{
    char buf[MAX_STRING_LENGTH];
    AppDataMCbrServer *serverPtr;

    ctoa(getSimTime(node), buf);
    clocktype t;

    switch (msg->eventType)
    {
        case MSG_APP_FromTransport:
        {
            UdpToAppRecv *info;
            MCbrData data;

            info = (UdpToAppRecv *) MESSAGE_ReturnInfo(msg);
            memcpy(&data, MESSAGE_ReturnPacket(msg), sizeof(data));

#ifdef DEBUG
            {
                char clockStr[24];
                ctoa(data.txTime, clockStr);
                printf("MCBR Server %u: packet transmitted at %s\n",
                       node->nodeId, clockStr);

                ctoa(getSimTime(node), clockStr);
                printf("    received at %s\n", clockStr);
                IO_ConvertIpAddressToString(&info->sourceAddr, addrStr);
                printf("    client is %s\n", addrStr);
                printf("    connection Id is %d\n", data.sourcePort);
                printf("    seqNo is %d\n", data.seqNo);
            }
#endif /* DEBUG */

            serverPtr = AppMCbrServerGetMCbrServer(node,
                                                   info->sourceAddr,
                                                   data.sourcePort);

            /* New connection, so create new mcbr server to handle client. */
            if (serverPtr == NULL)
            {
                serverPtr = AppMCbrServerNewMCbrServer(node,
                                                       info->destAddr,
                                                       info->sourceAddr,
                                                       data.sourcePort);
            }

            if (serverPtr == NULL)
            {
                printf("MCBR Server: Node %u unable to "
                       "allocation server\n", node->nodeId);

                assert(FALSE);
            }

            if (data.seqNo >= serverPtr->seqNo)
            {
                serverPtr->numBytesRecvd += MESSAGE_ReturnPacketSize(msg);
                serverPtr->numPktsRecvd++;
                serverPtr->sessionLastReceived = getSimTime(node);
                serverPtr->totalEndToEndDelay += getSimTime(node)
                                                 - data.txTime;

				t = getSimTime(node) - data.txTime;
				if(t > serverPtr->delay)
					serverPtr->totalDelayJitter += t - serverPtr->delay;
				else
					serverPtr->totalDelayJitter += serverPtr->delay - t;
				
				serverPtr->delay = t;

                serverPtr->seqNo = data.seqNo + 1;

#ifdef DEBUG
                {
                    char clockStr[24];
                    ctoa(serverPtr->totalEndToEndDelay, clockStr);
                    printf("    total end-to-end delay so far is %s\n",
                           clockStr);
                }
#endif /* DEBUG */

                if (data.type == 'd')
                {
                    /* Do nothing. */
                }
                else if (data.type == 'c')
                {
                    serverPtr->sessionFinish = getSimTime(node);
                    serverPtr->sessionIsClosed = TRUE;
                }
                else
                {
                    assert(FALSE);
                }
            }

            break;
        }

        default:
            ctoa(getSimTime(node), buf);
            printf("MCBR Server: At time %s, node %u received "
                   "message of unknown type "
                   "%u.\n", buf, node->nodeId, msg->eventType);
            assert(FALSE);
    }

    MESSAGE_Free(node, msg);
}

/*
 * NAME:        AppMCbrServerInit.
 * PURPOSE:     listen on MCbrServer server port.
 * PARAMETERS:  node - pointer to the node.
 * RETURN:      none.
 */
void
AppMCbrServerInit(Node *node)
{
}

/*
 * NAME:        AppMCbrServerPrintStats.
 * PURPOSE:     Prints statistics of a MCbrServer session.
 * PARAMETERS:  node - pointer to this node.
 *              serverPtr - pointer to the mcbr server data structure.
 * RETURN:      none.
 */
void
AppMCbrServerPrintStats(Node *node, AppDataMCbrServer *serverPtr) {
    clocktype throughput;

    char addrStr[MAX_STRING_LENGTH];
    char clockStr[MAX_STRING_LENGTH], clockStr2[MAX_STRING_LENGTH];
    char startStr[MAX_STRING_LENGTH];
    char closeStr[MAX_STRING_LENGTH];
    char sessionStatusStr[MAX_STRING_LENGTH];
    char throughputStr[MAX_STRING_LENGTH];
    char numByteRecvdStr[MAX_STRING_LENGTH];
    char buf[MAX_STRING_LENGTH];

    TIME_PrintClockInSecond(serverPtr->sessionStart, startStr);
    TIME_PrintClockInSecond(serverPtr->sessionLastReceived, closeStr);

    if (serverPtr->sessionIsClosed == FALSE) {
        serverPtr->sessionFinish = getSimTime(node);
        sprintf(sessionStatusStr, "Not closed");
    }
    else {
        sprintf(sessionStatusStr, "Closed");
    }

    if (serverPtr->numPktsRecvd == 0) {
        TIME_PrintClockInSecond(0, clockStr);
        TIME_PrintClockInSecond(0, clockStr2);
    }
    else {
        TIME_PrintClockInSecond(
            serverPtr->totalEndToEndDelay / serverPtr->numPktsRecvd,
            clockStr);
		TIME_PrintClockInSecond(
            serverPtr->totalDelayJitter / serverPtr->numPktsRecvd,
            clockStr2);
    }

    if (serverPtr->sessionFinish <= serverPtr->sessionStart) {
        throughput = 0;
    }
    else {
        throughput = (clocktype)
                     ((clocktype)(serverPtr->numBytesRecvd) * SECOND
                     / (serverPtr->sessionFinish - serverPtr->sessionStart));

        throughput = throughput * 8;
    }

    ctoa(throughput, throughputStr);
    ctoa(serverPtr->numBytesRecvd, numByteRecvdStr);

    IO_ConvertIpAddressToString(&serverPtr->remoteAddr, addrStr);

    sprintf(buf, "Client Address = %s", addrStr);
    IO_PrintStat(
        node,
        "Application",
        "MCBR Server",
        ANY_DEST,
        serverPtr->sourcePort,
        buf);

    sprintf(buf, "First Packet Received at (s) = %s", startStr);
    IO_PrintStat(
        node,
        "Application",
        "MCBR Server",
        ANY_DEST,
        serverPtr->sourcePort,
        buf);

    sprintf(buf, "Last Packet Received at (s) = %s", closeStr);
    IO_PrintStat(
        node,
        "Application",
        "MCBR Server",
        ANY_DEST,
        serverPtr->sourcePort,
        buf);

	sprintf(buf, "Average Delay Jitter (s) = %s", clockStr2);
    IO_PrintStat(
        node,
        "Application",
        "MCBR Server",
        ANY_DEST,
        serverPtr->sourcePort,
        buf);

    sprintf(buf, "Average End-to-End Delay (s) = %s", clockStr);
    IO_PrintStat(
        node,
        "Application",
        "MCBR Server",
        ANY_DEST,
        serverPtr->sourcePort,
        buf);

    sprintf(buf, "Session Status = %s", sessionStatusStr);
    IO_PrintStat(
        node,
        "Application",
        "MCBR Server",
        ANY_DEST,
        serverPtr->sourcePort,
        buf);


    sprintf(buf, "Total Bytes Received = %s", numByteRecvdStr);
    IO_PrintStat(
        node,
        "Application",
        "MCBR Server",
        ANY_DEST,
        serverPtr->sourcePort,
        buf);

    sprintf(buf, "Total Packets Received = %d", serverPtr->numPktsRecvd);
    IO_PrintStat(
        node,
        "Application",
        "MCBR Server",
        ANY_DEST,
        serverPtr->sourcePort,
        buf);

    sprintf(buf, "Throughput (bits/s) = %s", throughputStr);
    IO_PrintStat(
        node,
        "Application",
        "MCBR Server",
        ANY_DEST,
        serverPtr->sourcePort,
        buf);
}

/*
 * NAME:        AppMCbrServerFinalize.
 * PURPOSE:     Collect statistics of a MCbrServer session.
 * PARAMETERS:  node - pointer to the node.
 *              appInfo - pointer to the application info data structure.
 * RETURN:      none.
 */
void
AppMCbrServerFinalize(Node *node, AppInfo* appInfo)
{
    AppDataMCbrServer *serverPtr = (AppDataMCbrServer*)appInfo->appDetail;

    if (node->appData.appStats == TRUE)
    {
        AppMCbrServerPrintStats(node, serverPtr);
    }
}

/*
 * NAME:        AppMCbrServerGetMCbrServer.
 * PURPOSE:     search for a mcbr server data structure.
 * PARAMETERS:  node - mcbr server.
 *              remoteAddr - mcbr client sending the data.
 *              sourcePort - sourcePort of mcbr client/server pair.
 * RETURN:      the pointer to the mcbr server data structure,
 *              NULL if nothing found.
 */
AppDataMCbrServer * //inline//
AppMCbrServerGetMCbrServer(Node *node, Address remoteAddr, int sourcePort)
{
    AppInfo *appList = node->appData.appPtr;
    AppDataMCbrServer *mcbrServer;

    for (; appList != NULL; appList = appList->appNext)
    {
        if (appList->appType == APP_MCBR_SERVER)
        {
            mcbrServer = (AppDataMCbrServer *) appList->appDetail;

            if (mcbrServer->sourcePort == sourcePort
                && Address_IsSameAddress(&mcbrServer->remoteAddr, &remoteAddr))
            {
                return mcbrServer;
            }
       }
   }

    return NULL;
}

/*
 * NAME:        AppMCbrServerNewMCbrServer.
 * PURPOSE:     create a new mcbr server data structure, place it
 *              at the beginning of the application list.
 * PARAMETERS:  node - pointer to the node.
 *              localAddr - local address.
 *              remoteAddr - remote address.
 *              sourcePort - sourcePort of mcbr client/server pair.
 * RETRUN:      the pointer to the created mcbr server data structure,
 *              NULL if no data structure allocated.
 */
AppDataMCbrServer * //inline//
AppMCbrServerNewMCbrServer(Node *node,
                           Address localAddr,
                           Address remoteAddr,
                           int sourcePort)
{
    AppDataMCbrServer *mcbrServer;

    mcbrServer = (AppDataMCbrServer *)
                MEM_malloc(sizeof(AppDataMCbrServer));

    /*
     * Fill in mcbr info.
     */
    memcpy(&(mcbrServer->localAddr), &localAddr, sizeof(Address));
    memcpy(&(mcbrServer->remoteAddr), &remoteAddr, sizeof(Address));
    mcbrServer->sourcePort = sourcePort;
    mcbrServer->sessionStart = getSimTime(node);
    mcbrServer->sessionFinish = getSimTime(node);
    mcbrServer->sessionLastReceived = getSimTime(node);
    mcbrServer->sessionIsClosed = FALSE;
    mcbrServer->numBytesRecvd = 0;
    mcbrServer->numPktsRecvd = 0;
    mcbrServer->totalEndToEndDelay = 0;
    mcbrServer->seqNo = 0;

	mcbrServer->totalDelayJitter = 0;
	mcbrServer->delay = 0;

    APP_RegisterNewApp(node, APP_MCBR_SERVER, mcbrServer);

    return mcbrServer;
}

