// Copyright (c) 2001-2006, Scalable Network Technologies, Inc.  All Rights Reserved.
//                          6701 Center Drive West
//                          Suite 520
//                          Los Angeles, CA 90045
//                          sales@scalable-networks.com
//
// This source code is licensed, not sold, and is subject to a written
// license agreement.  Among other things, no portion of this source
// code may be copied, transmitted, disclosed, displayed, distributed,
// translated, used as the basis for a derivative work, or used, in
// whole or in part, for any program or purpose other than its intended
// use in compliance with the license agreement as part of the QualNet
// software.  This source code and certain of the algorithms contained
// within it are confidential trade secrets of Scalable Network
// Technologies, Inc. and may not be used as the basis for any other
// software, hardware, product or service.

#ifndef MCBR_APP_H
#define MCBR_APP_H

#include "types.h"

/*
 * Data item structure used by multicast cbr.
 */
typedef
struct struct_app_mcbr_data
{
    Int32 sourcePort;
    char type;
    Int32 seqNo;
    clocktype txTime;
}
MCbrData;

/* Structure containing mcbr client information. */

typedef
struct struct_app_mcbr_client_str
{
    Address localAddr;
    Address remoteAddr;
    clocktype interval;
    clocktype sessionStart;
    clocktype sessionFinish;
    clocktype sessionLastSent;
    clocktype endTime;
    BOOL sessionIsClosed;
    Int64 numBytesSent;
    Int32 numPktsSent;
    Int32 itemsToSend;
    Int32 itemSize;
    Int32 sourcePort;
    Int32 seqNo;
}
AppDataMCbrClient;

/* Structure containing mcbr related information. */

typedef
struct struct_app_mcbr_server_str
{
    Address localAddr;
    Address remoteAddr;
    int sourcePort;
    clocktype sessionStart;
    clocktype sessionFinish;
    clocktype sessionLastReceived;
    BOOL sessionIsClosed;
    Int64 numBytesRecvd;
    Int32 numPktsRecvd;
    clocktype totalEndToEndDelay, totalDelayJitter, delay;  //Modified by Marvin
    Int32 seqNo;
}
AppDataMCbrServer;

/*
 * NAME:        AppLayerMCbrClient.
 * PURPOSE:     Models the behaviour of MCbrClient Client on receiving the
 *              message encapsulated in msg.
 * PARAMETERS:  nodePtr - pointer to the node which received the message.
 *              msg - message received by the layer
 * RETURN:      none.
 */
void
AppLayerMCbrClient(Node *nodePtr, Message *msg);

/*
 * NAME:        AppMCbrClientInit.
 * PURPOSE:     Initialize a MCbrClient session.
 * PARAMETERS:  nodePtr - pointer to the node,
 *              serverAddr - address of the server,
 *              itemsToSend - number of items to send,
 *              itemSize - size of each packet,
 *              interval - interval of packet transmission rate.
 *              startTime - time until the session starts.
 *              endTime - time until the session end.
 * RETURN:      none.
 */
void
AppMCbrClientInit(
    Node *nodePtr,
    Address clientAddr,
    Address serverAddr,
    Int32 itemsToSend,
    Int32 itemSize,
    clocktype interval,
    clocktype startTime,
    clocktype endTime);

/*
 * NAME:        AppMCbrClientPrintStats.
 * PURPOSE:     Prints statistics of a MCbrClient session.
 * PARAMETERS:  nodePtr - pointer to the this node.
 *              clientPtr - pointer to the mcbr client data structure.
 * RETURN:      none.
 */
void
AppMCbrClientPrintStats(Node *nodePtr, AppDataMCbrClient *clientPtr);

/*
 * NAME:        AppMCbrClientFinalize.
 * PURPOSE:     Collect statistics of a MCbrClient session.
 * PARAMETERS:  nodePtr - pointer to the node.
 *              appInfo - pointer to the application info data structure.
 * RETURN:      none.
 */
void
AppMCbrClientFinalize(Node *nodePtr, AppInfo* appInfo);

/*
 * NAME:        AppMCbrClientGetMCbrClient.
 * PURPOSE:     search for a mcbr client data structure.
 * PARAMETERS:  nodePtr - pointer to the node.
 *              sourcePort - connection ID of the mcbr client.
 * RETURN:      the pointer to the mcbr client data structure,
 *              NULL if nothing found.
 */
AppDataMCbrClient * //inline//
AppMCbrClientGetMCbrClient(Node *nodePtr, int sourcePort);

/*
 * NAME:        AppMCbrClientNewMCbrClient.
 * PURPOSE:     create a new mcbr client data structure, place it
 *              at the beginning of the application list.
 * PARAMETERS:  nodePtr - pointer to the node.
 *              remoteAddr - remote address.
 *              itemsToSend - number of items to send,
 *              itemSize - size of data packets.
 *              interval - interdeparture time of packets.
 *              startTime - time when session is started.
 *              endTime - time when session is ended.
 * RETURN:      the pointer to the created mcbr client data structure,
 *              NULL if no data structure allocated.
 */
AppDataMCbrClient * //inline//
AppMCbrClientNewMCbrClient(
    Node *nodePtr,
    Address localAddr,
    Address remoteAddr,
    Int32 itemsToSend,
    Int32 itemSize,
    clocktype interval,
    clocktype startTime,
    clocktype endTime);

/*
 * NAME:        AppMCbrClientScheduleNextPkt.
 * PURPOSE:     schedule the next packet the client will send.  If next packet
 *              won't arrive until the finish deadline, schedule a close.
 * PARAMETERS:  nodePtr - pointer to the node,
 *              clientPtr - pointer to the mcbr client data structure.
 * RETRUN:      none.
 */
void //inline//
AppMCbrClientScheduleNextPkt(Node *nodePtr, AppDataMCbrClient *clientPtr);



/*
 * NAME:        AppLayerMCbrServer.
 * PURPOSE:     Models the behaviour of MCbrServer Server on receiving the
 *              message encapsulated in msg.
 * PARAMETERS:  nodePtr - pointer to the node which received the message.
 *              msg - message received by the layer
 * RETURN:      none.
 */
void
AppLayerMCbrServer(Node *nodePtr, Message *msg);

/*
 * NAME:        AppMCbrServerInit.
 * PURPOSE:     listen on MCbrServer server port.
 * PARAMETERS:  nodePtr - pointer to the node.
 * RETURN:      none.
 */
void
AppMCbrServerInit(Node *nodePtr);

/*
 * NAME:        AppMCbrServerPrintStats.
 * PURPOSE:     Prints statistics of a MCbrServer session.
 * PARAMETERS:  nodePtr - pointer to this node.
 *              serverPtr - pointer to the mcbr server data structure.
 * RETURN:      none.
 */
void
AppMCbrServerPrintStats(Node *nodePtr, AppDataMCbrServer *serverPtr);

/*
 * NAME:        AppMCbrServerFinalize.
 * PURPOSE:     Collect statistics of a MCbrServer session.
 * PARAMETERS:  nodePtr - pointer to the node.
 *              appInfo - pointer to the application info data structure.
 * RETURN:      none.
 */
void
AppMCbrServerFinalize(Node *nodePtr, AppInfo* appInfo);

/*
 * NAME:        AppMCbrServerGetMCbrServer.
 * PURPOSE:     search for a mcbr server data structure.
 * PARAMETERS:  nodePtr - mcbr server.
 *              sourcePort - sourcePort of mcbr client/server pair.
 *              remoteAddr - mcbr client sending the data.
 * RETURN:      the pointer to the mcbr server data structure,
 *              NULL if nothing found.
 */
AppDataMCbrServer * //inline//
AppMCbrServerGetMCbrServer(
    Node *nodePtr,
    Address remoteAddr,
    int sourcePort);

/*
 * NAME:        AppMCbrServerNewMCbrServer.
 * PURPOSE:     create a new mcbr server data structure, place it
 *              at the beginning of the application list.
 * PARAMETERS:  nodePtr - pointer to the node.
 *              localAddr - local address.
 *              remoteAddr - remote address.
 *              sourcePort - sourcePort of mcbr client/server pair.
 * RETRUN:      the pointer to the created mcbr server data structure,
 *              NULL if no data structure allocated.
 */
AppDataMCbrServer * //inline//
AppMCbrServerNewMCbrServer(
    Node *nodePtr,
    Address localAddr,
    Address remoteAddr,
    int sourcePort);

#endif /* _MCBR_APP_H_ */
