/*
filename: multicast_sptm.cpp
last modified: April 28, 2008
author: Marvin Pinto
email: marvin@expr.ca
*/


/* Include Directives */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "random.h"
#include "api.h"
#include "clock.h"
#include "partition.h"
#include "scheduler.h"
#include "network.h"
#include "node.h"
#include "network_ip.h"



#include "multicast_sptm.h"

char initialNode = TRUE;    //for the absolutely first node the init function gets called for
SptmData *spt;  // data structure for state variables (only created once)
int finalizeNodeCtr;    //used for memory cleanup purposed in the finalize function



/*
This is the SPTM initialization function called by QualNet.

For the very first node initialized, it will allocate the protocol data structure
along with the memory and initialize all the nodes in the simulation.

*/
void SptmInit(Node* node,
            const NodeInput* nodeInput,
            int interfaceIndex)
{

    #if SPTM_DEBUG
    printf("SptmInit: Attempting to check if it's already instantiated\n");
    #endif


    /* Initialize state variables only for the first node */
    if (initialNode)
    {

        #if SPTM_DEBUG
        printf("SptmInit: Begin Protocol Initialization\n");
        #endif
    
    	// Allocate memory for the SPTM data structure 
    	spt = (SptmData *)MEM_malloc (sizeof(SptmData));

    	
    	// Initialize to 0, will be incremented in the finalize function
    	finalizeNodeCtr = 0;
    
    
        #if SPTM_DEBUG
        printf("SptmInit: Memory for SptmData allocated\n");
        #endif
    
    
    	// Read the relevant parameters from the config file
    	//   and set the state variables accordingly.
    	SptmInitializeConfigurableParameters(node,nodeInput);
    
        #if SPTM_DEBUG
        printf("SptmInit: SptmInitializeConfigurableParameters Complete\n");
        #endif
    
    
    
        // Initialize sptm data structure variables 
        spt->simulationStats.avgHopCount = 0.0;
        spt->simulationStats.totalNumForwarders = 0;
        spt->simulationStats.totalSenders = 0;
        spt->simulationStats.totalReceivers = 0;
        spt->simulationStats.configuredSenders = 0;
        spt->simulationStats.configuredReceivers = 0;
        spt->processHello = TRUE;
    
        // For our purposes, "infinity" will be considered to be ULONG_MAX
        spt->infty = ULONG_MAX-1;
        
        // Initialize sequenceNumber to 1
        spt->sequenceNumber = 1;
    
   
        #if SPTM_DEBUG
        printf("SptmInit: Maxrange is currently set to %d\n", spt->maxRange);
        printf("SptmInit: Infinity is currently set to %d\n", spt->infty);
        printf("SptmInit: Initialization of data structure variables Complete\n");
        #endif

        // Initialize all the nodes in the graph
        InitializeAllGraphNodes(node,interfaceIndex);
        

        #if SPTM_DEBUG
        printf("SptmInit: InitializeAllGraphNodes Complete\n");
        #endif
        
        // Read in the multicast senders from the APP-CONFIG-FILE parameter
        ReadMulticastSendersFromInputFile(nodeInput);
    
        #if SPTM_DEBUG
        printf("SptmInit: read multicast senders Complete\n");
        #endif

        // Schedule the broadcast of the MSG_NETWORK_CheckTimeoutAlarm msg
        //   This message will be used to trigger the stoppage of the hello
        //   messages, and the creation of the Dijkstra Trees
        clocktype delay = (clocktype)(spt->helloDuration * SPTM_HELLO_DURATION);
        SptmSetTimer(node, MSG_NETWORK_CheckTimeoutAlarm,delay);
    
        #if SPTM_DEBUG
        char clockString[MAX_STRING_LENGTH];
        TIME_PrintClockInSecond(delay, clockString);
        printf("SptmInit: MSG_NETWORK_CheckTimeoutAlarm message scheduled\n");
        printf("SptmInit: scheduled for %s seconds\n", clockString);
        #endif

        initialNode=FALSE;
    } //END IF 

	// Set the multicast routing protocol, and the router
	//   function to SPTM.
	NetworkIpSetMulticastRoutingProtocol(node, spt, interfaceIndex);
	NetworkIpSetMulticastRouterFunction(node, &SptmRouterFunction, interfaceIndex);

    #if SPTM_DEBUG
    printf("SptmInit: Qualnet pointers set\n");
    #endif
    

    RandomSeed randomSeed;
    RANDOM_SetSeed(randomSeed, node->globalSeed, node->nodeId, MULTICAST_PROTOCOL_SPTM);

    #if SPTM_DEBUG
    printf("SptmInit: Random seeds set\n");
    #endif

    // Schedule the broadcasts of hello messages
    // delay range: 0 to 1 second
    RANDOM_SetSeed(randomSeed, node->globalSeed, node->nodeId, MULTICAST_PROTOCOL_SPTM);
    clocktype delay = (clocktype)( RANDOM_erand(randomSeed) * SPTM_HELLO_INTERVAL );
    SptmSetTimer(node, MSG_NETWORK_SendHello, delay);


    #if SPTM_DEBUG
    printf("SptmInit: Hello messages scheduled\n");
    #endif


} /* END SptmInit Function */








/*
    This function will read the file specified in the APP-CONFIG-FILE 
    paramater (config file), and parse it for the multicast senders.
*/
static void ReadMulticastSendersFromInputFile(const NodeInput *nodeInput)
{
    NodeInput mcastInput;
    BOOL retVal;

    #if SPTM_DEBUG
    printf("ReadMcastSndrs: Function Call\n");
    #endif
    
    IO_ReadCachedFile(ANY_NODEID, ANY_ADDRESS, nodeInput,
                         "APP-CONFIG-FILE", &retVal,
                         &mcastInput);

    #if SPTM_DEBUG
    printf("ReadMcastSndrs: File read in\n");
    #endif

    InitializeList(&(spt->mcastGroups));

    #if SPTM_DEBUG
    printf("ReadMcastSndrs: Multicast group list initialized\n");
    #endif

    for (int i=0;i<mcastInput.numLines;i++)
    {

        #if SPTM_DEBUG
        printf("ReadMcastSndrs: Line %d of mcast input file\n", i);
        #endif

        //misc variables
        char appType[MAX_STRING_LENGTH];    //MCBR, CBR, etc.
        NodeAddress srcAddr, mcastAddr;
        char mcastString[MAX_STRING_LENGTH];

        #if SPTM_DEBUG
        printf("ReadMcastSndrs: variables set\n");
        #endif

        // Read line 'i'
        sscanf(mcastInput.inputStrings[i],
                "%s %u %s %*s %*s %*s %*s %*s",
                appType, &srcAddr, mcastString);

        #if SPTM_DEBUG
        printf("ReadMcastSndrs: line %d read in\n", i);
        #endif

        // e.g. "MCBR 1 225.0.0.1 0 512 0.25S 200S 800S"
        // Only process if it's a multicast application string
        if ( strcmp(appType, "MCBR")==0 )
        {
            int numHostBits;
            BOOL isNodeId;

            SptmMcastGrp* newGroup = (SptmMcastGrp *)malloc( sizeof(SptmMcastGrp) );

            #if SPTM_DEBUG
            printf("ReadMcastSndrs: new group mem allocated\n");
            #endif

            IO_ParseNodeIdHostOrNetworkAddress(mcastString, &mcastAddr, &numHostBits, &isNodeId);

            #if SPTM_DEBUG
            printf("ReadMcastSndrs: network address parsed\n");
            #endif

            newGroup->id = mcastAddr;
            newGroup->sender = AssociateWithSptmNode(srcAddr);
            newGroup->sender->nodeStats.isNodeASender = TRUE;
            spt->simulationStats.totalSenders++;

            InitializeList(&(newGroup->receivers));
            char retVal = AddToList( (void*)newGroup, &(spt->mcastGroups), "SptmMcastGrp" );


            #if SPTM_DEBUG
            printf("ReadMcastSndrs: Add group %s with sender %u to multicast group list\n",
                    mcastString, srcAddr);
            #endif

        } //END IF
    } //END FOR LOOP
} //END ReadMulticastSendersFromInputFile FUNCTION















/*
    This function will be called automatically by Qualnet.
    Qualnet essentially reads the multicast group file and "joins" the respective
    group at the time point set in the multicast group file.
    
    sends MSG_NETWORK_JoinGroup with delay set in the file
*/
void SptmJoinGroup(Node *node, NodeAddress mcastAddr)
{
    #if SPTM_DEBUG_L2
    printf("SptmJoinGroup: Function call\n");
    #endif


    SptmNode *sptNode = AssociateWithSptmNode(node);
    
    
    // Search through all the multicast groups
    SptmLinkedListItem *curr = spt->mcastGroups.head.next; //first mcast group
    for (int i=0;i<spt->mcastGroups.size;i++)
    {

        // Add this receiver to the appropriate mcast group
        if (  ((SptmMcastGrp*)(curr->data))->id == mcastAddr  )
        {
            AddToList( (void*)sptNode, &(((SptmMcastGrp*)(curr->data))->receivers), "SptmMcastGrp" );
            break;
        } //END IF

        curr = curr->next;
    } //END FOR LOOP
    
    sptNode->nodeStats.joinQueriesSent++;
    sptNode->nodeStats.isNodeAReceiver = TRUE;
    spt->simulationStats.totalReceivers++;

    #if SPTM_DEBUG_L2
    char addressStr[MAX_STRING_LENGTH];
    IO_ConvertIpAddressToString(mcastAddr, addressStr);
    printf("SptmJoinGroup: Node %u joined group %s\n", sptNode->id, addressStr);
    #endif

} // End SptmJoinGroup Function






/*
    This function will be called automatically by Qualnet.
    Qualnet essentially reads the multicast group file and "leaves" the respective
    group at the time point set in the multicast group file.
    
    sends MSG_NETWORK_LeaveGroup with delay set in the file
*/
void SptmLeaveGroup(Node *node, NodeAddress mcastAddr)
{
    #if SPTM_DEBUG
    printf("SptmLeaveGroup: Function call\n");
    #endif
    
    SptmNode *sptNode = AssociateWithSptmNode(node);

    #if SPTM_DEBUG
    char addressStr[MAX_STRING_LENGTH];
    IO_ConvertIpAddressToString(mcastAddr, addressStr);
    printf("SptmLeaveGroup: Node %d leaves group %s\n", sptNode->id, addressStr);
    #endif
    
    sptNode->nodeStats.leaveQueriesSent++;
    
    // Search through all the multicast groups
    SptmLinkedListItem *curr = spt->mcastGroups.head.next; //first mcast group
    for (int i=0;i<GetItemCount(&(spt->mcastGroups));i++)
    {

        // Add this receiver to the appropriate mcast group
        if (  ((SptmMcastGrp*)(curr->data))->id == mcastAddr  )
        {
            curr = RemoveFromList(
                            (void*)sptNode, 
                            &(((SptmMcastGrp*)(curr->data))->receivers),
                            "SptmNode");
            //free(curr);
            break;
        } //END IF

        curr = curr->next;
    } //END FOR LOOP
    free(curr);
} //END SptmLeaveGroup FUNCTION








/*
    This function is automatically called by Qualnet.  It is the core function
    that takes care of all the routing.
*/
void SptmRouterFunction(Node* node,
                        Message* msg,
                        NodeAddress destAddr,
                        int interfaceIndex,
                        BOOL* packetWasRouted,
                        NodeAddress prevHop)
{
    #if SPTM_DEBUG_L2
    printf("SptmRouterFunction: Function Call\n");
    #endif

    SptmNode *sptNode = AssociateWithSptmNode(node);

    IpHeaderType* ipHeader = (IpHeaderType *)MESSAGE_ReturnPacket(msg);
    NodeAddress msgSource = ipHeader->ip_src;
    NodeAddress msgDest = ipHeader->ip_dst;
    clocktype delay;
    
    *packetWasRouted = TRUE;

    // Only process if this is a UDP MCBR msg
    if(ipHeader->ip_p == IPPROTO_UDP)
    {

        //In order to determine if this node is a forwarder
        //for this particular multicast group
        SptmLinkedListItem* fwdForGrp = IsItemInList(&msgDest, 
                                                     &(sptNode->forwarderForGroup), 
                                                     "SptmMcastGrp");
        #if SPTM_DEBUG_L3
        char addressStr[MAX_STRING_LENGTH];
        //IO_ConvertIpAddressToString( msgDest, addressStr );
        //printf("SptmRouterFunction:node %u, sender %u, group %s\n", sptNode->id, msgSource, addressStr);
        #endif
        
        // If this is a multicast sender
        if (msgSource == sptNode->id)
        {
            #if SPTM_DEBUG_L2
            printf("SptmRouterFunction: Node %u is a sender\n", sptNode->id);
            #endif
            
            // Add the sequence option to the ip header
            SptmIpHeaderOption opt;
            opt.seqNo = GetUniqueSequenceNumber();
            AddIpOptionField(node, msg, IPOPT_SPTM, sizeof(SptmIpHeaderOption));
            SptmSetIpOption(msg, &opt);

            // Send the message            
            sptNode->nodeStats.numPacketsSent++;
            NetworkIpSendPacketToMacLayer(node, 
                                            msg, 
                                            DEFAULT_INTERFACE, 
                                            ANY_DEST);

            //*packetWasRouted = TRUE;
        } //END IF
        // This is just another node along the way, and if it's a forwarder
        else if ( sptNode->nodeStats.isNodeAForwarder == TRUE &&
                  fwdForGrp!=NULL)
        {
        
            // Get the sequence number from the ip header
            SptmIpHeaderOption opt = SptmGetIpHeaderOption(msg);
            SptmLinkedListItem* nodeCache = IsItemInList(&msgSource, 
                                                         &(sptNode->msgCache), 
                                                         "SptmMessageCache");
            // First time this sender is encountered, add it to the list
            if (nodeCache == NULL)
            {
                SptmMessageCache *cacheItem = 
                        (SptmMessageCache *)MEM_malloc(sizeof(SptmMessageCache));
                cacheItem->id = msgSource;
                cacheItem->seqNo = opt.seqNo;
                AddToList( (void*)cacheItem, &(sptNode->msgCache), "SptmMessageCache");
                
                
                // Proceed to forward the packet onwards
                sptNode->nodeStats.numPacketsForwarded++;
                sptNode->nodeStats.numPacketsReceived++;
                RandomSeed randomSeed;
                RANDOM_SetSeed(randomSeed, node->globalSeed, node->nodeId, MULTICAST_PROTOCOL_SPTM);
                // delay range: 0 to n milliseconds (n being the value read from the config file)
                delay = (clocktype)(RANDOM_erand(randomSeed) * spt->delayJitter * SPTM_DELAY_JITTER);

                #if SPTM_DEBUG_L3
                //char addressStr[MAX_STRING_LENGTH];
                IO_ConvertIpAddressToString( msgDest, addressStr );
                printf("SptmRouterFunction:node %u is a forwarder for %s\n", sptNode->id, addressStr);
                #endif
                
                NetworkIpSendPacketToMacLayerWithDelay(node, 
                                                        msg, 
                                                        DEFAULT_INTERFACE, 
                                                        ANY_DEST,
                                                        delay);
                //*packetWasRouted = TRUE;
            } //END IF nodecache==null
            else 
            {
                /*
                    This sender was already in the list.  Proceed to check that the
                    last sequence number in the cache is less than the current
                    sequence number.  Otherwise, dump the packet, it's a duplicate.
                */
            
                // This is good, not a duplicate packet
                if (  ((SptmMessageCache*)(nodeCache->data))->seqNo < opt.seqNo  )
                {
                    ((SptmMessageCache*)(nodeCache->data))->seqNo = opt.seqNo;
                    
                    // Proceed to forward the packet onwards
                    sptNode->nodeStats.numPacketsForwarded++;
                    sptNode->nodeStats.numPacketsReceived++;
                    RandomSeed randomSeed;
                    RANDOM_SetSeed(randomSeed, 
                                   node->globalSeed, 
                                   node->nodeId, 
                                   MULTICAST_PROTOCOL_SPTM);
                    // delay range: 0 to n milliseconds
                    delay = (clocktype)(RANDOM_erand(randomSeed) * 
                                        spt->delayJitter * SPTM_DELAY_JITTER);

                    #if SPTM_DEBUG_L3
                    //char addressStr[MAX_STRING_LENGTH];
                    IO_ConvertIpAddressToString( msgDest, addressStr );
                    printf("SptmRouterFunction:node %u is a forwarder for %s\n", sptNode->id, addressStr);
                    #endif
                    
                    NetworkIpSendPacketToMacLayerWithDelay(node, 
                                                            msg, 
                                                            DEFAULT_INTERFACE, 
                                                            ANY_DEST,
                                                            delay);
                    //*packetWasRouted = TRUE;
                } //END IF


                else    // Proceed to dump the packet
                {
                    //*packetWasRouted = FALSE;
                    sptNode->nodeStats.numPacketsDropped++;
                    MESSAGE_Free(node,msg);
                } //END ELSE
            
            } //END ELSE
        } //END ELSE IF
        else
        {
            //*packetWasRouted = FALSE;
            sptNode->nodeStats.numPacketsDropped++;
            MESSAGE_Free(node,msg);
        } //END ELSE
    } //END IF IPPROTO_UDP
    else
        MESSAGE_Free(node,msg);

} //END SptmRouterFunction Function




/*
    Extract the custom ip header option from the message
*/
static SptmIpHeaderOption SptmGetIpHeaderOption(const Message* msg)
{
    IpOptionsHeaderType* ipOption =
        FindAnIpOptionField((IpHeaderType*)msg->packet, IPOPT_SPTM);
		
    SptmIpHeaderOption opt;
	
    memcpy(&opt,
        ((char*)ipOption + sizeof(IpOptionsHeaderType)),
        sizeof(SptmIpHeaderOption) );
    return opt;
} //END SptmGetIpHeaderOption FUNCTION





/*
    Set the custom ip header option on a message
*/
static void SptmSetIpOption(Message* msg, const SptmIpHeaderOption* opt)
{
    IpOptionsHeaderType* ipOption =
        FindAnIpOptionField((IpHeaderType*)msg->packet, IPOPT_SPTM);
		
    memcpy(((char*)ipOption + sizeof(IpOptionsHeaderType)),
        opt,
        sizeof(SptmIpHeaderOption) );
} //END SptmSetIpOption FUNCTION










static void SptmSetTimer(Node* node, int eventType, clocktype delay)
{

    #if SPTM_DEBUG
    printf("SptmSetTimer: Function call\n");
    #endif

    Message* newMsg = NULL;
    Address* info = NULL;
    NetworkRoutingProtocolType protocolType;

    protocolType = MULTICAST_PROTOCOL_SPTM;


    // Broadcast address
    Address destAddr;
    destAddr.networkType = NETWORK_IPV4;
    destAddr.interfaceAddr.ipv4 = ANY_DEST;


    // Allocate message for the timer
    newMsg = MESSAGE_Alloc(
                 node,
                 NETWORK_LAYER,
                 protocolType,
                 eventType);

    // Assign the address for which the timer is meant for
    MESSAGE_InfoAlloc(
        node,
        newMsg,
        sizeof(Address));

    info = (Address *) MESSAGE_ReturnInfo(newMsg);

    memcpy(info, &destAddr, sizeof(Address));

    // Schedule the timer after the specified delay
    MESSAGE_Send(node, newMsg, delay);
    
    
    #if SPTM_DEBUG
    char clockString[MAX_STRING_LENGTH];
    TIME_PrintClockInSecond(delay, clockString);
    printf("SptmSetTimer: Node: %d, Event: %d, Delay: %s\n", 
                node->nodeId, eventType, clockString);
    #endif
    
} //END SptmSetTimer FUNCTION


















static void SptmInitializeConfigurableParameters(
				Node* node, const NodeInput* nodeInput)
{
    #if SPTM_DEBUG
    printf("SptmInitializeConfigurableParameters: Function call \n");
    #endif

    BOOL retVal;
    char buf[MAX_STRING_LENGTH];

	// Determine whether or not to collect routing statistics
	// This parameter is read from the config file
	IO_ReadString(node->nodeId, ANY_ADDRESS,
	   		      nodeInput,"ROUTING-STATISTICS",&retVal,buf);
	if ((retVal == FALSE) || (strcmp(buf, "NO") == 0))
	{ spt->collectStats = FALSE; }
	else if (strcmp(buf, "YES") == 0)
	{ spt->collectStats = TRUE; }


    //unsigned long readVal;
    int readVal;
	// Read the SPTM-MAXRANGE value from the config file
	// Defaults to 350m if no range is specified
	IO_ReadInt(node->nodeId, ANY_ADDRESS,
	   		      nodeInput,"SPTM-MAXRANGE",&retVal,&readVal);
	if ((retVal == FALSE)) { spt->maxRange = 350; }
	else { spt->maxRange = (unsigned long)readVal; }



    int readVal2;
	// Read the SPTM-HELLO-DURATION value from the config file
	// Defaults to 75 seconds if no value is specified
	IO_ReadInt(node->nodeId, ANY_ADDRESS,
	   		      nodeInput,"SPTM-HELLO-DURATION",&retVal,&readVal2);
	if ((retVal == FALSE)) { spt->helloDuration = 75; }
	else { spt->helloDuration = readVal2; }


	// Read the SPTM-MAX-DELAY-JITTER value from the config file
	// Defaults to 5 milliseconds if no value is specified
	IO_ReadInt(node->nodeId, ANY_ADDRESS,
	   		      nodeInput,"SPTM-MAX-DELAY-JITTER",&retVal,&readVal2);
	if ((retVal == FALSE)) { spt->delayJitter = 5; }
	else { spt->delayJitter = readVal2; }


    #if SPTM_DEBUG
    printf("SptmInitializeConfigurableParameters: Function complete \n");
    #endif

} //END SptmInitializeConfigurableParameters FUNCTION
















void SptmHandleProtocolEvent(Node* node, Message* msg)
{

    #if SPTM_DEBUG
    printf("SptmHandleProtocolEvent: Function call \n");
    #endif

	SptmHello hello;
	Message* hmsg;
	clocktype delay;

    SptmNode* sptNode = AssociateWithSptmNode(node);

	switch (msg->eventType)
	{

        // Case: Trigger to terminate all the hello msgs, and begin
        // creation of the Dijkstra trees.
		case MSG_NETWORK_CheckTimeoutAlarm:
		{
            #if SPTM_DEBUG
            printf("SptmHandleProtocolEvent: Case MSG_NETWORK_CheckTimeoutAlarm\n");
            printf("SptmHandleProtocolEvent: processHello is %d\n", spt->processHello);
            #endif
		    // This needs to be done only once
		    if ( (spt->processHello)==TRUE )
		    {
		        // Cease all further hello messages
		        spt->processHello = FALSE;

                // Build the Dijkstra shortest path tree(s) rooted from the senders
                BuildDijkstraTrees(node);


                // Print out all the multicast groups
                //PrintMulticastGroups();


                #if SPTM_DEBUG
                printf("SptmHandleProtocolEvent: Dijkstra trees complete\n");
                #endif
		    } //END IF
		    break;
		} //END CASE



		// Case: A hello timer has expired and needs to be sent
		case MSG_NETWORK_SendHello:
		{
            #if SPTM_DEBUG
            printf("SptmHandleProtocolEvent: Case MSG_NETWORK_SendHello\n");
            #endif
		    // If the hello timer hasn't gone off, process this hello msg
		    if ( (spt->processHello)==TRUE )
		    {
    			// Allocate enough memory for the hello message
    			// and proceed to broadcast it.
    			hello.sendingNode = AssociateWithSptmNode(node);
    			hmsg = MESSAGE_Alloc(node,MAC_LAYER,0,MSG_MAC_FromNetwork);
    			MESSAGE_PacketAlloc(node,hmsg,sizeof(SptmHello),TRACE_SPTM);
    			memcpy((char *)(hmsg->packet),&hello,sizeof(SptmHello) );

    			NetworkIpSendRawMessageToMacLayer(node,
            		hmsg,
            		NetworkIpGetInterfaceAddress(node, DEFAULT_INTERFACE),
            		ANY_DEST,
            		IPTOS_PREC_INTERNETCONTROL,
            		IPPROTO_SPTM,
            		1,
            		DEFAULT_INTERFACE,
            		ANY_DEST);

    			sptNode->nodeStats.numHelloSent++;

                RandomSeed randomSeed;
                RANDOM_SetSeed(randomSeed, node->globalSeed, node->nodeId, MULTICAST_PROTOCOL_SPTM);
    			// delay range: 0 to 1 second
    			delay = (clocktype)(RANDOM_erand(randomSeed) * SPTM_HELLO_INTERVAL);
    			SptmSetTimer(node, MSG_NETWORK_SendHello, delay);
                #if SPTM_DEBUG
                printf("SptmHandleProtocolEvent: Node %d sent a hello msg\n", node->nodeId);
                #endif
			} //END IF
			break;
		} //END CASE
		default:
		{ printf("Error: Node %d handle protocol event: unknown event\n", node->nodeId); }
	} //END SWITCH STATEMENT

	MESSAGE_Free(node, msg);
    #if SPTM_DEBUG
    printf("SptmHandleProtocolEvent: Function complete\n");
    #endif
} //END SptmHandleProtocolEvent Function









static void PrintMulticastGroups(SptmMcastGrp* group)
{
    char addressStr[MAX_STRING_LENGTH];
    //SptmLinkedListItem* curr = spt->mcastGroups.head.next;  // First mcast group
    
    printf("\n\n");
    
    // Do the following for each multicast group
    //for (int i=0;i<GetItemCount(&(spt->mcastGroups));i++)
    //{
        //int numReceivers = GetItemCount(  &(((SptmMcastGrp*)(curr->data))->receivers)  );
        int numReceivers = GetItemCount( &(group->receivers) );
        //IO_ConvertIpAddressToString( ((SptmMcastGrp*)(curr->data))->id, addressStr );
        IO_ConvertIpAddressToString( group->id, addressStr );
        printf("Multicast Group %s (%d receivers)\n", addressStr, numReceivers);
        printf("Multicast Sender %u\n", group->sender->id);


        // Do the following for each receiver of this group
        SptmLinkedListItem* curr2 = group->receivers.head.next;
        for (int x=0;x<numReceivers;x++)
        {
            printf("Multicast Receiver %u\n", ((SptmNode*)(curr2->data))->id );
            SptmNode* curr3 = ((SptmNode*)(curr2->data));
            while ( curr3->id != group->sender->id )
            {
                curr3 = curr3->prevNode;
                if (curr3 == NULL)  // case where the receiver is "stranded" out of range
                    break;
                else
                    printf("\tPrev path: %u\n", curr3->id);
            } //END WHILE LOOP
            curr2 = curr2->next;    //next receiver
        } //END FOR LOOP
        printf("\n");
        
        //curr = curr->next;  //next mcast group
    //} //END FOR LOOP
    printf("\n\n");
} //END PrintMulticastGroups FUNCTION












void SptmHandleProtocolPacket(Node* node, Message* msg)
{
	// General Design: after node initialization, all the nodes will
	// broadcast hello messages for a fixed period of time, after which
	// a trigger will be sent to calculate the shortest path trees.
	// At this point, all hello messages will cease to be sent (or processed)
	// and for any nodes for which hello wasn't processed, it will be
	// assumed that they are not within transmission range, and will
	// therefore be ignored.

    #if SPTM_DEBUG
    printf("SptmHandleProtocolPacket: Function call\n");
    #endif


	int size = MESSAGE_ReturnPacketSize(msg);
	SptmNode *sender, *receiver;

	// If a hello message is received
	if(size == sizeof(SptmHello))
	{


		//If this node receives a hello message, and the transmitter is also within maxRange,
		//	then update the "adjacent nodes" list of this node.
		sender = ((SptmHello *)(msg->packet))->sendingNode;
		receiver = AssociateWithSptmNode(node);
		
		receiver->nodeStats.numHelloReceived++;

        #if SPTM_DEBUG
        printf("SptmHandleProtocolPacket: Node %u received a hello msg from node %u\n", 
                                        node->nodeId, sender->id);
        printf("SptmHandleProtocolPacket: Dist between Node %u and node %u is %u\n", 
                                        node->nodeId, sender->id,
                                        CalculateDistance(sender, receiver)
                                        );
        #endif

        // The sender can potentially be added to the list iff it's within maxRange.

		if (CalculateDistance(sender, receiver) <=  spt->maxRange )
		{
		    //Add the sender to the adjacent nodes list iff it's not already in there
		    AddToList( (void*)sender, &(receiver->adjacentNodes), "SptmNode" );
		    
            #if SPTM_DEBUG
            printf("SptmHandleProtocolPacket: node %d added node %d to its adjacent nodes list\n",
                            receiver->id, sender->id);
            #endif

		} //END IF

		MESSAGE_Free(node, msg);

	}
	else
	{
		printf("Error: Node %d handle protocol packet: unknown packet\n", node->nodeId);
		MESSAGE_Free(node, msg);
	}

    #if SPTM_DEBUG
    printf("SptmHandleProtocolPacket: Function complete\n");
    #endif
} //END SptmHandleProtocolPacket Function


















/*
	This function takes a nodeid as a parameter
	and returns the corresponding SptmNode.
*/
static SptmNode* AssociateWithSptmNode(NodeAddress nodeId)
{
    #if SPTM_DEBUG
    printf("AssociateWithSptmNode nodeId: Function call\n");
    printf("AssociateWithSptmNode nodeId: item count: %d\n", GetItemCount(&(spt->sptmNodes)));
    
    #endif

    

    SptmLinkedListItem *curr = spt->sptmNodes.head.next;
    for (int i=0;i<GetItemCount(&(spt->sptmNodes));i++)
    {
        if (  ((SptmNode*)(curr->data))->id == nodeId )
        {
            #if SPTM_DEBUG
            printf("AssociateWithSptmNode nodeId: nodeId = %u\n", nodeId);
            printf("AssociateWithSptmNode nodeId: spt.nodeId = %u\n", ((SptmNode*)(curr->data))->id);
            #endif
            return (SptmNode*)(curr->data);
        } //END IF
        curr = curr->next;
    } //END FOR LOOP

	return NULL;
} //END AssociateWithSptmNode FUNCTION








/*
	This function takes the generic Node data type as a parameter
	and returns the corresponding SptmNode.
*/
static SptmNode* AssociateWithSptmNode(Node *node)
{
    #if SPTM_DEBUG
    printf("AssociateWithSptmNode *node: Function call\n");
    printf("AssociateWithSptmNode *node: item count: %d\n", GetItemCount(&(spt->sptmNodes)));
    #endif

    SptmLinkedListItem *curr = spt->sptmNodes.head.next;
    for (int i=0;i<GetItemCount(&(spt->sptmNodes));i++)
    {
        if (  ((SptmNode*)(curr->data))->id ==  node->nodeId )
        {
            #if SPTM_DEBUG
            printf("AssociateWithSptmNode *node: nodeId = %u\n", node->nodeId);
            printf("AssociateWithSptmNode *node: spt.nodeId = %u\n", ((SptmNode*)(curr->data))->id);
            #endif
            return (SptmNode*)(curr->data);
        } //END IF
        curr = curr->next;
    } //END FOR LOOP

	return NULL;
} //END AssociateWithSptmNode FUNCTION



















static void InitializeAllGraphNodes(Node* node, int interfaceIndex)
{
    #if SPTM_DEBUG
    printf("InitializeAllGraphNodes: Function call\n");
    #endif

    // Initialize sptmNodes linked list
    InitializeList(&(spt->sptmNodes));
    

    #if SPTM_DEBUG
    printf("InitializeAllGraphNodes: Memory for linked list allocated\n");
    #endif

    

    /* Initialize the data structure variables for all the nodes */
    Node *x = node->partitionData->firstNode;
    for (int i=0;i<(node->numNodes);i++)
    {
        #if SPTM_DEBUG
        printf("InitializeAllGraphNodes: Initializing DS for node %d\n", x->nodeId);
        #endif

        SptmNode* newNode = (SptmNode *)malloc( sizeof(SptmNode) );
        newNode->id = x->nodeId;
        newNode->weight = spt->infty;
        newNode->isCovered = FALSE;
        newNode->prevNode = NULL;
        
        // Initialize Node Stats
        newNode->nodeStats.isNodeAForwarder = FALSE;
    	newNode->nodeStats.hopCount = 0;
    	newNode->nodeStats.numHelloSent = 0;
    	newNode->nodeStats.numHelloReceived = 0;
    	newNode->nodeStats.joinQueriesSent = 0;
    	newNode->nodeStats.leaveQueriesSent = 0;
    	newNode->nodeStats.numPacketsReceived = 0;
    	newNode->nodeStats.numPacketsForwarded = 0;
    	newNode->nodeStats.numPacketsDropped = 0;
    	newNode->nodeStats.isNodeASender = FALSE;
    	newNode->nodeStats.numPacketsSent = 0;
    	newNode->nodeStats.isNodeAReceiver = FALSE;


        // Initialize the message cache data structure for this node
        InitializeList(&(newNode->msgCache));
        

        #if SPTM_DEBUG
        printf("InitializeAllGraphNodes: Prelim variables init\n");
        #endif

        newNode->xcoord = x->mobilityData->current->position.common.c1;
        newNode->ycoord = x->mobilityData->current->position.common.c2;
        newNode->zcoord = x->mobilityData->current->position.common.c3;

        #if SPTM_DEBUG
        printf("InitializeAllGraphNodes: Coordinates init\n");
        #endif

        AddToList( (void*)newNode, &(spt->sptmNodes), "SptmNode" );

        #if SPTM_DEBUG
        printf("InitializeAllGraphNodes: node added to list\n");
        #endif

        InitializeList(&(newNode->adjacentNodes));
        InitializeList(&(newNode->forwarderForGroup));

        #if SPTM_DEBUG
        printf("InitializeAllGraphNodes: Node %d - weight %d \n",
                newNode->id, newNode->weight );
        printf("InitializeAllGraphNodes: Node %d - isCovered %d \n",
                newNode->id, newNode->isCovered );
        printf("InitializeAllGraphNodes: Node %d - forwarder %d \n",
                newNode->id, newNode->nodeStats.isNodeAForwarder );
        printf("InitializeAllGraphNodes: Node %d - XYZ %lf %lf %lf \n",
                newNode->id, newNode->xcoord, newNode->ycoord, newNode->zcoord );
        #endif


        x = x->nextNodeData;
    } //END FOR LOOP

    #if SPTM_DEBUG
    printf("InitializeAllGraphNodes: Function complete\n");
    #endif
} //END InitializeAllGraphNodes Function






/*
    Function will return the next number in order.
    The underlying assumption here is that if there are too many packets
    transmitted by the senders, then overflow error
    limit: approx 4.2 billion
*/
static unsigned long GetUniqueSequenceNumber()
{
    unsigned long x = spt->sequenceNumber;
    spt->sequenceNumber++;
	return x;	
} //END GetUniqueSequenceNumber FUNCTION










/*
    - If a node are designated as a forwarder by Dijkstra, any
    packets that it receives will be forwarded through its wireless
    interface to ANY_DEST (i.e. broadcast)


    Essentially, the procedure will be as follows:
        - Starting from a sender, build a shortest path tree rooted from it
          to the rest of the graph.
        - After the tree is built, traverse back from the receivers and
          begin to mark the nodes as forwarders.
        - Repeat this procedure for each multicast sender.
*/
static void BuildDijkstraTrees(Node* node)
{
    #if SPTM_DEBUG_L2
    printf("BuildDijkstraTrees: Function call\n");
    #endif
    
    // # of multicast groups (i.e. trees)
    int numTrees = GetItemCount(&(spt->mcastGroups));
    
    // Traverse the adjacent nodes list for each sender, updating the
    // node weight, and adding them to the nodesToCover list.
    // We will now also calculate the hopcount for each
    // tree (rooted from each sender).
    float totalHC = 0;
    
    SptmLinkedListItem *curr = spt->mcastGroups.head.next; //First mcast group
    
    // Do the following for each multicast group
    for (int i=0; i < numTrees; i++)
    {

        // Update the node weights using Dijkstra's algorithm
        UpdateNodeWeights( ((SptmMcastGrp*)(curr->data))->sender );

        #if SPTM_DEBUG_L2
        printf("BuildDijkstraTrees: Node weights updated for sender %u\n", 
                                ((SptmMcastGrp*)(curr->data))->sender->id);
        #endif


        // Mark all the forwarding nodes for this mcast group
        // traversing back from the receivers.
        totalHC = totalHC + MarkForwardingNodes(  (SptmMcastGrp*)(curr->data)  );

        PrintMulticastGroups((SptmMcastGrp*)(curr->data));
        curr=curr->next;    //next mcast group
    } //END FOR LOOP

    // Count all the forwarding nodes, senders, receivers
    SumSimulationStats();

    // Determine average hopcount for these trees
    spt->simulationStats.avgHopCount = totalHC/numTrees;

    #if SPTM_DEBUG_L2
    printf("BuildDijkstraTrees: TotalHC: %f, Trials: %d, AvgHC: %f\n", 
                            totalHC, numTrees, spt->simulationStats.avgHopCount);
    printf("BuildDijkstraTrees: Total Mcast Fowarders: %d\n", spt->simulationStats.totalNumForwarders);
    printf("BuildDijkstraTrees: Function complete \n");
    #endif

} //END BuildDijkstraTree Function






/*
    Function to sum up some relevant simulation stats
*/
static void SumSimulationStats()
{
    SptmLinkedListItem* currNode = spt->sptmNodes.head.next; //First node
    int numNodes = GetItemCount( &(spt->sptmNodes) );   //total nodes
    for (int i=0;i<numNodes;i++)
    {
        if ( ((SptmNode*)(currNode->data))->nodeStats.isNodeAForwarder == TRUE )
            spt->simulationStats.totalNumForwarders++;

        if ( ((SptmNode*)(currNode->data))->nodeStats.isNodeASender == TRUE )
            spt->simulationStats.configuredSenders++;

        if ( ((SptmNode*)(currNode->data))->nodeStats.isNodeAReceiver == TRUE )
            spt->simulationStats.configuredReceivers++;

        currNode = currNode->next;
    } //END FOR LOOP
} //END SumSimulationStats FUNCTION






/*
    FUNCTION   : MarkForwardingNodes
    PURPOSE    : After UpdateNodeWeights has been ran, traverse backwards from
                 the receivers (using the prevNode pointer) to mark all the nodes
                 along the way as forwarding nodes (including the sender).
    PARAMETERS :
                 +SptmNode:*sender: Pointer to the specified sender
                 +SptmNode:*receiver: Pointer to the specified receiver
    RETURN     : int:currHopCount: The hopcount of the path between the sender
                                   and the receiver.
    NOTES      : A receiver will be marked as a forwarding node iff it is part
                 of another sender/receiver tree which needs this node to
                 forward data.  By default, a receiver is NOT marked as a
                 forwarding node.
*/
static float MarkForwardingNodes(SptmMcastGrp* group)
{
    #if SPTM_DEBUG_L2
    printf("MarkForwardingNodes: Function call \n");
    #endif
    
    int totalHC=0;
   

    int numReceivers = GetItemCount(&(group->receivers));
    SptmLinkedListItem *currRecv = group->receivers.head.next; //First receiver


    #if SPTM_DEBUG_L2
    printf("MarkForwardingNodes: number of receivers %d \n", numReceivers);
    #endif

    // Do the following for each receiver
    for (int i=0;i<numReceivers;i++)
    {

        #if SPTM_DEBUG_L2
        printf("MarkForwardingNodes: now working on receiver %u \n", 
                            ((SptmNode*)(currRecv->data))->id );
        #endif
    
        SptmNode* tempRecv = (SptmNode*)(currRecv->data);   //current receiver
        
        #if SPTM_DEBUG_L2
        if (tempRecv->prevNode == NULL)
            printf("MarkForwardingNodes: tempRecv->prevnode == NULL \n");
        printf("MarkForwardingNodes: tempRecv variable assigned \n");
        #endif
        
        // Traverse back up the path for this receiver
        while (0<1)
        {
        
            #if SPTM_DEBUG_L2
            printf("MarkForwardingNodes: prevnode id is %u, sender is %u\n",
                                    tempRecv->prevNode->id,
                                    group->sender->id);
            #endif

            
            // This is the scenario where this receiver is completely out
            // of range.  i.e. its nearest node is too far away to 
            // communicate with it, and therefore there is no path to it
            // via the dijkstra tree.
            if (tempRecv->prevNode == NULL)
            {
                spt->simulationStats.totalReceivers--;
                break;
            } //END IF


            // If the previous node is the sender
            if (  tempRecv->prevNode->id  == group->sender->id  )
            {
                tempRecv->prevNode->nodeStats.isNodeAForwarder = TRUE;
                totalHC++;
                char retVal = AddToList( (void*)group,
                                         &(tempRecv->prevNode->forwarderForGroup),
                                         "SptmMcastGrp" );

                #if SPTM_DEBUG_L2
                printf("MarkForwardingNodes: prev node is sender %u \n", 
                                    tempRecv->prevNode->id );
                #endif
                
                // update the HC for our receiver
                ((SptmNode*)(currRecv->data))->nodeStats.hopCount++;
                break;
            } //END IF
            else    // If the previous node is anyone but the sender
            {
                tempRecv->prevNode->nodeStats.isNodeAForwarder = TRUE;
                totalHC++;
                char retVal = AddToList( (void*)group,
                                         &(tempRecv->prevNode->forwarderForGroup),
                                         "SptmMcastGrp" );

                #if SPTM_DEBUG_L2
                printf("MarkForwardingNodes: prev node is forwarder %u \n", 
                                    tempRecv->prevNode->id );
                #endif

                tempRecv = tempRecv->prevNode;                
                // update the HC for our receiver
                ((SptmNode*)(currRecv->data))->nodeStats.hopCount++;
            } //END ELSE
        } //END WHILE LOOP
    
        currRecv = currRecv->next;  // Next receiver
    } //END FOR LOOP


    #if SPTM_DEBUG_L2
    printf("MarkForwardingNodes: function done, about to return HC \n");
    #endif

    // Calculate and return the average hopcount of these trees
    return (float)totalHC / numReceivers;
} //END FUNCTION



















/*
    FUNCTION   : UpdateNodeWeights
    PURPOSE    : Use Dijkstra's algorithm to update all the node weights of
                 the graph starting from the sender.
    PARAMETERS :
                 +Node:*node: Pointer to the a generic Qualnet node
                 +SptmNode:*sender: Pointer to the specified sender
    RETURN     : VOID
    NOTES      : Instead of implementing a "heap" of sorts to store the node
                 weights, a simple linked list will be used with which the minimum
                 value will be searched for each time.  This is a potentially
                 expensive operation but applicable to our scenario
                 (relatively small # of nodes).
*/
static void UpdateNodeWeights(SptmNode *sender)
{
    #if SPTM_DEBUG_L2
    printf("UpdateNodeWeights: Function call \n");
    #endif


    /* First reinitialize the weight, isCovered, and prevNode state
       variables for each node of the graph. */
    SptmLinkedListItem *curr = spt->sptmNodes.head.next;
    for (int i=0;i<GetItemCount(&(spt->sptmNodes));i++)
    {
        ((SptmNode*)(curr->data))->weight = spt->infty;
        ((SptmNode*)(curr->data))->isCovered = FALSE;
        ((SptmNode*)(curr->data))->prevNode = NULL;
        curr=curr->next;
    } /* END FOR LOOP */

    #if SPTM_DEBUG_L2
    printf("UpdateNodeWeights: nodes reinitialized \n");
    #endif

    // Linked list to store the nodes that are not covered
    SptmLinkedList *nodesToCover = (SptmLinkedList *)malloc( sizeof(SptmLinkedList) );
    InitializeList(nodesToCover);

    #if SPTM_DEBUG_L2
    printf("UpdateNodeWeights: nodesToCover initialized \n");
    #endif    
    
    // Add the sender to the nodesToCover list
    sender->weight = 0;     //Dijkstra
    AddToList( (void*)sender, nodesToCover, "SptmNode");

    #if SPTM_DEBUG_L2
    printf("UpdateNodeWeights: sender added to nodesToCover \n");
    #endif


    // This is the situation where the sender is "stranded" too far
    // from any nearby node
    if ( GetItemCount(nodesToCover)==0 )
        spt->simulationStats.totalSenders--;
    


    // Proceed to loop through all the nodes in the nodesToCover list:
    // Selecting the node with the smallest weight, update all its immediate neighbours
    // edge weights and add them to the nodesToCover list if they're not already in there.
    // Note: This is done iff isCovered==FALSE
    while (GetItemCount(nodesToCover) > 0)
    {

        #if SPTM_DEBUG_L2
        printf("UpdateNodeWeights: nodesToCover size: %d \n", GetItemCount(nodesToCover));
        #endif
        
        // Search through the list for the node with the smallest weight
        unsigned long minSize = spt->infty;
        
        #if SPTM_DEBUG_L2
        printf("UpdateNodeWeights: minsize: %d \n", minSize);
        #endif
        
        SptmLinkedListItem *curr, *winner;
        curr = nodesToCover->head.next;    // First item in the list
        for (int i=0;i<GetItemCount(nodesToCover);i++)
        {
            if ( ((SptmNode*)(curr->data))->weight <= minSize )
            {
                winner = curr;
                minSize = ((SptmNode*)(curr->data))->weight;
            } //END IF
            curr=curr->next;
        } //END FOR LOOP
        //"winner" is therefore the item with the smallest weight

        #if SPTM_DEBUG_L2
        printf("UpdateNodeWeights: winner is node %u \n", ((SptmNode*)(winner->data))->id);
        #endif


        // Now update all "winner"'s adjacent nodes with the weight,
        // which also implies that the prevNode will be updated if
        // weight is changed.  All these adjacent nodes will also be added
        // to the nodesToCover list iff their boolean value isCovered==FALSE
        curr = ((SptmNode*)(winner->data))->adjacentNodes.head.next;  //first item
        int itemCount = GetItemCount(&( ((SptmNode*)(winner->data))->adjacentNodes ));
        for (int i=0; i <itemCount; i++)
        {

            // Calculate the distance between the nodes, only process if within "range"
            unsigned long dist;
            dist = CalculateDistance( (SptmNode*)(curr->data), (SptmNode*)(winner->data) );

            #if SPTM_DEBUG_L2
            printf("UpdateNodeWeights: distance between nodes %u and %u is %d\n", 
                        ((SptmNode*)(curr->data))->id,
                        ((SptmNode*)(winner->data))->id,
                        dist);
            #endif

            // If node is within our specified range AND node is not covered AND
            // the weight of the child node is >= weight of the parent + 1 (one hop away),
            // then process
            // In order to change this logic to actual distance calculation (rather than one hop),
            // assign tempDist = dist
            unsigned long tempDist = 1;
            if ( dist<=(spt->maxRange)  &&
                 ((SptmNode*)(curr->data))->isCovered == FALSE &&
                 ((SptmNode*)(curr->data))->weight >= ((SptmNode*)(winner->data))->weight + tempDist
               )
            {
                ((SptmNode*)(curr->data))->weight = ((SptmNode*)(winner->data))->weight + tempDist;
                ((SptmNode*)(curr->data))->prevNode = (SptmNode*)(winner->data);
                
                // Add this node to the nodesToCover list
                AddToList( (void*)(  (SptmNode*)(curr->data)  ),  nodesToCover, "SptmNode");

                #if SPTM_DEBUG_L2
                printf("UpdateNodeWeights:1 node %u has prevNode node %u \n", 
                                            ((SptmNode*)(curr->data))->id,
                                            ((SptmNode*)(curr->data))->prevNode->id);
                #endif

            
            } //END IF

            #if SPTM_DEBUG_L2
            printf("UpdateNodeWeights:1 nodesToCover size is %d \n", GetItemCount(nodesToCover));
            #endif
            
            curr = curr->next;
        } //END FOR LOOP



        // Now remove the "winner" from the nodesToCover list
        // and set it as a "covered" node.
        ((SptmNode*)(winner->data))->isCovered = TRUE;
        curr = RemoveFromList(
                        (SptmNode*)(winner->data),  
                        nodesToCover,
                        "SptmNode");
        free(curr);

        #if SPTM_DEBUG_L2
        printf("UpdateNodeWeights:2 nodesToCover size is %d \n", GetItemCount(nodesToCover));
        #endif

    } //END WHILE LOOP

    #if SPTM_DEBUG_L2
    printf("UpdateNodeWeights: about to free nodesToCover \n");
    #endif

    // Cleanup
    free(nodesToCover);

    #if SPTM_DEBUG_L2
    printf("UpdateNodeWeights: nodesToCover has been freed \n");
    #endif

} //END FUNCTION







/*
    This function searches the linked list for the specified item
    and returns a pointer to the list item (without deleting it from the list),
    or returns NULL if it isn't present.
*/
static SptmLinkedListItem* IsItemInList( void* itemToSearch, SptmLinkedList* list, char* type)
{
    #if SPTM_DEBUG
    printf("IsItemInList: Function Call \n");
    #endif

    SptmLinkedListItem* curr = list->head.next;

    #if SPTM_DEBUG
    printf("IsItemInList: curr pointer \n");
    printf("IsItemInList: type is %s \n", type);
    #endif


    /*
        The following if statements were only put in place in order to have
        a sort of "generic" linked list data structure.
    */
    if (strcmp(type,"SptmNode")==0)
    {
        #if SPTM_DEBUG
        printf("IsItemInList: Therefore at SptmNode \n");
        printf("IsItemInList: list size = %d \n", list->size);
        #endif
   
        for (int i=0; i<(list->size); i++)
        {
            #if SPTM_DEBUG
            printf("IsItemInList: comparing to item %d \n", i);
            #endif
            if (  ((SptmNode *)itemToSearch)->id == ((SptmNode *)(curr->data))->id  )
                return curr;
            curr = curr->next;
        } //END FOR LOOP    
    } //END IF
    
    
    else if (strcmp(type,"SptmMcastGrp")==0)
    {
        #if SPTM_DEBUG
        printf("IsItemInList: Therefore at SptmMcastGrp \n");
        printf("IsItemInList: list size = %d \n", list->size);
        #endif

        for (int i=0; i<(list->size); i++)
        {
            if (  ((SptmMcastGrp *)itemToSearch)->id == ((SptmMcastGrp *)(curr->data))->id  )
                return curr;
            curr = curr->next;
        } //END FOR LOOP
    } //END ELSE IF


    else if (strcmp(type,"SptmMessageCache")==0)
    {
        #if SPTM_DEBUG
        printf("IsItemInList: Therefore at SptmMessageCache \n");
        printf("IsItemInList: list size = %d \n", list->size);
        #endif

        for (int i=0; i<(list->size); i++)
        {
            if (  ((SptmMessageCache *)itemToSearch)->id == ((SptmMessageCache *)(curr->data))->id  )
                return curr;
            curr = curr->next;
        } //END FOR LOOP
    } //END ELSE IF


    return NULL;
} //END IsItemInList FUNCTION




/*
    Overloaded function with generic id type as the item to search
*/
static SptmLinkedListItem* IsItemInList( NodeAddress* itemToSearch, SptmLinkedList* list, char* type)
{
    #if SPTM_DEBUG
    printf("IsItemInList: Function Call \n");
    #endif

    SptmLinkedListItem* curr = list->head.next;

    #if SPTM_DEBUG
    printf("IsItemInList: curr pointer \n");
    printf("IsItemInList: type is %s \n", type);
    #endif


    /*
        The following if statements were only put in place in order to have
        a sort of "generic" linked list data structure.
    */
    if (strcmp(type,"SptmNode")==0)
    {
        #if SPTM_DEBUG
        printf("IsItemInList: Therefore at SptmNode \n");
        printf("IsItemInList: list size = %d \n", list->size);
        #endif
   
        for (int i=0; i<(list->size); i++)
        {
            #if SPTM_DEBUG
            printf("IsItemInList: comparing to item %d \n", i);
            #endif
            
            if (  *itemToSearch == ((SptmNode *)(curr->data))->id  )
                return curr;
            curr = curr->next;
        } //END FOR LOOP    
    } //END IF
    
    
    else if (strcmp(type,"SptmMcastGrp")==0)
    {
        #if SPTM_DEBUG
        printf("IsItemInList: Therefore at SptmMcastGrp \n");
        printf("IsItemInList: list size = %d \n", list->size);
        #endif

        for (int i=0; i<(list->size); i++)
        {
            if (  *itemToSearch == ((SptmMcastGrp *)(curr->data))->id  )
                return curr;
            curr = curr->next;
        } //END FOR LOOP
    } //END ELSE IF


    else if (strcmp(type,"SptmMessageCache")==0)
    {
        #if SPTM_DEBUG
        printf("IsItemInList: Therefore at SptmMessageCache \n");
        printf("IsItemInList: list size = %d \n", list->size);
        #endif

        for (int i=0; i<(list->size); i++)
        {
            if (  *itemToSearch == ((SptmMessageCache *)(curr->data))->id  )
                return curr;
            curr = curr->next;
        } //END FOR LOOP
    } //END ELSE IF



    return NULL;
} //END IsItemInList FUNCTION








static void InitializeList(SptmLinkedList* list)
{
    list->size=0;
    list->head.prev = NULL;
    list->head.next = &(list->tail);
    list->head.data = NULL;
    list->tail.prev = &(list->head);
    list->tail.next = NULL;
    list->tail.data = NULL;
} //END InitializeList FUNCTION








/*
    This function searches the linked list for the specified item and
    proceeds to remove it if it's in there.  The calling function is responsible
    for freeing the associated memory.
    
    Returns the specified item to be deleted or NULL
*/
static SptmLinkedListItem* RemoveFromList( void* itemToRemove, 
                                           SptmLinkedList* list,
                                           char* type)
{
    SptmLinkedListItem* temp = IsItemInList(itemToRemove, list, type);
    if (temp != NULL)
    {
        #if SPTM_DEBUG
        printf("RemoveFromList: This item IS present in this list \n");
        #endif
    
        SptmLinkedListItem* before = temp->prev;
        SptmLinkedListItem* after = temp->next;
        before->next = after;
        after->prev = before;
        list->size--;

        #if SPTM_DEBUG
        printf("RemoveFromList: Tasks completed \n");
        #endif
        
    } //END IF
    
    return temp;
} //END RemoveFromList FUNCTION







/*
    This function adds the specified item to the linked list
    iff it's not already in there.
    
    Returns TRUE or FALSE depending on success or failure.
    FALSE implies the item was already in the list.
*/
static char AddToList( void* itemToAdd, SptmLinkedList* list, char* type)
{
    #if SPTM_DEBUG
    printf("AddToList: Function Call \n");
    #endif

    char retVal = FALSE;

    #if SPTM_DEBUG
    printf("AddToList: retval = false \n");
    #endif


    SptmLinkedListItem *temp = IsItemInList(itemToAdd, list, type);
    
    #if SPTM_DEBUG
    printf("AddToList: temp initialized \n");
    #endif

    if (temp == NULL)
    {
        #if SPTM_DEBUG
        printf("AddToList: Item is NOT in the list \n");
        #endif    
    
        SptmLinkedListItem *newItem = (SptmLinkedListItem *)malloc( sizeof(SptmLinkedListItem) );

        #if SPTM_DEBUG
        printf("AddToList: allocate memory \n");
        #endif    

      
        newItem->data = itemToAdd;
        newItem->next = &(list->tail);
        newItem->prev = list->tail.prev;

        #if SPTM_DEBUG
        printf("AddToList: set newitem pointers \n");
        #endif    
        
        newItem->prev->next = newItem;
        list->tail.prev = newItem;

        #if SPTM_DEBUG
        printf("AddToList: set prev pointers \n");
        #endif    

        list->size++;
        
        #if SPTM_DEBUG
        printf("AddToList: list size %d\n", list->size);
        #endif    
        
        retVal = TRUE;
    } //END IF

    return retVal;
} //END AddToList FUNCTION








/*
    This function returns the number of items in the
    specified linked list.
*/
static int GetItemCount( SptmLinkedList* list)
{
    return list->size;
} //END GetItemCount FUNCTION















/*
    FUNCTION   : CalculateDistance
    PURPOSE    : Calculate and return the distance between the
                 two specified nodes.
    PARAMETERS :
                 +SptmNode:*n1:Pointer to the first node
                 +SptmNode:*n2:Pointer to the second node
    RETURN     : double: the distance between the nodes
*/
static unsigned long CalculateDistance(SptmNode *n1, SptmNode *n2)
{
	return (unsigned long)sqrt( (n2->xcoord-n1->xcoord)*(n2->xcoord-n1->xcoord) +
	             (n2->ycoord-n1->ycoord)*(n2->ycoord-n1->ycoord) +
	             (n2->zcoord-n1->zcoord)*(n2->zcoord-n1->zcoord) );
} //END CalculateDistance Function















/*
    Print out all the relvant simulation/node stats and begin freeing up
    all memory used.
*/
void SptmFinalize(Node *node)
{
    #if SPTM_DEBUG
    printf("SptmFinalize: Function call\n");
    #endif 

    // Print out all the relevant simulation stats if warranted
    if (spt->collectStats == TRUE)
        PrintStats(node);

    #if SPTM_DEBUG
    printf("SptmFinalize: Statistics Printed\n");
    #endif 


    // Free the memory used for the message cache data structure
    SptmNode* thisNode = AssociateWithSptmNode(node);
    for (int i=0;i<GetItemCount(&(thisNode->msgCache));i++)
    {
        SptmLinkedListItem *curr;
        SptmLinkedListItem *curr2;
        
        curr = thisNode->msgCache.head.next; //Current Item
        curr2 = RemoveFromList(curr->data, &(thisNode->msgCache), "SptmMessageCache");
        free(curr2->data);
        free(curr2);
    } //END FOR LOOP



    // When this value reaches the node limit (i.e. final node), memory
    // cleanup will commence.
    finalizeNodeCtr++;
    
    
    // If we've reached the final node, begin memory cleanup
    if ( finalizeNodeCtr >= GetItemCount( &(spt->sptmNodes) )  )
    {
        int numItems = 0;
        SptmLinkedListItem *curr;
        SptmLinkedListItem *curr2;
        
    
        // Free the memory used for all the multicast groups
        numItems = GetItemCount(&(spt->mcastGroups));
        for (int i=0;i<numItems;i++)
        {
            curr = spt->mcastGroups.head.next;   //Current item
            curr2 = RemoveFromList(curr->data, &(spt->mcastGroups), "SptmMcastGrp");
            
            #if SPTM_DEBUG
            printf("SptmFinalize:1 Attempting to free curr2->data\n");
            printf("SptmFinalize:1 curr2->data->id is %u\n", ((SptmMcastGrp*)(curr2->data))->id  );
            #endif
            
            free(curr2->data);
            
            #if SPTM_DEBUG
            printf("SptmFinalize:1 Attempting to free curr\n");
            #endif
            
            free(curr2);
        } //END FOR LOOP
    
        #if SPTM_DEBUG
        printf("SptmFinalize: Memory for mcast groups freed\n");
        #endif 
    
    
        // Free the memory used for all the nodes
        numItems = GetItemCount(&(spt->sptmNodes));
        for (int i=0;i<numItems;i++)
        {
            curr = spt->sptmNodes.head.next;   //Current item
            curr2 = RemoveFromList(curr->data, &(spt->sptmNodes), "SptmNode");
            
            #if SPTM_DEBUG
            printf("SptmFinalize:2 attempting to free curr2->data \n");
            #endif 
            
            free(curr2->data);
    
            #if SPTM_DEBUG
            printf("SptmFinalize:2 attempting to free curr2 \n");
            #endif 
    
            free(curr2);
        } //END FOR LOOP
    
    
        #if SPTM_DEBUG
        printf("SptmFinalize: Memory for sptmnodes freed\n");
        #endif 
    
        // Free the memory used for the spt data structure
        free(spt);
        
        #if SPTM_DEBUG
        printf("SptmFinalize: Memory for spt DS freed\n");
        #endif 
    } //END IF
    
} /* End SptmFinalize Function */







/*
This function will print the predefined statistics
to the output stat file.
*/
static void PrintStats(Node *node)
{
    char buf[MAX_STRING_LENGTH];

    // Print out simulation stats
    sprintf(buf, "Simulation - Average Hop Count = %f", spt->simulationStats.avgHopCount);
    IO_PrintStat(node,"Network","SPTM",ANY_DEST,-1,buf);


    sprintf(buf, "Simulation - Total Number of Forwarders = %i", 
                                    spt->simulationStats.totalNumForwarders);
    IO_PrintStat(node,"Network","SPTM",ANY_DEST,-1,buf);


    sprintf(buf, "Simulation - Total Number of actual Senders = %i", 
                                   spt->simulationStats.totalSenders);
    IO_PrintStat(node,"Network","SPTM",ANY_DEST,-1,buf);

   
    sprintf(buf, "Simulation - Total Number of actual Receivers = %i", 
                                    spt->simulationStats.totalReceivers);
    IO_PrintStat(node,"Network","SPTM",ANY_DEST,-1,buf);
    

    sprintf(buf, "Simulation - Total Number of configured Senders = %i", 
                                   spt->simulationStats.configuredSenders);
    IO_PrintStat(node,"Network","SPTM",ANY_DEST,-1,buf);

   
    sprintf(buf, "Simulation - Total Number of configured Receivers = %i", 
                                    spt->simulationStats.configuredReceivers);
    IO_PrintStat(node,"Network","SPTM",ANY_DEST,-1,buf);

    
    // Print out individual node stats
    SptmNode* sptNode = AssociateWithSptmNode(node);

    
    sprintf(buf, "Node %u - Hopcount = %i", sptNode->id, sptNode->nodeStats.hopCount);
    IO_PrintStat(node,"Network","SPTM",ANY_DEST,-1,buf);

    sprintf(buf, "Node %u - Is this node a forwarder = %d", sptNode->id, sptNode->nodeStats.isNodeAForwarder);
    IO_PrintStat(node,"Network","SPTM",ANY_DEST,-1,buf);


    sprintf(buf, "Node %u - Is this node a sender = %d", sptNode->id, sptNode->nodeStats.isNodeASender);
    IO_PrintStat(node,"Network","SPTM",ANY_DEST,-1,buf);


    sprintf(buf, "Node %u - Is this node a receiver = %d", sptNode->id, sptNode->nodeStats.isNodeAReceiver);
    IO_PrintStat(node,"Network","SPTM",ANY_DEST,-1,buf);


    sprintf(buf, "Node %u - Number hello's sent = %i", sptNode->id, sptNode->nodeStats.numHelloSent);
    IO_PrintStat(node,"Network","SPTM",ANY_DEST,-1,buf);


    sprintf(buf, "Node %u - Number hello's received = %i", sptNode->id, sptNode->nodeStats.numHelloReceived);
    IO_PrintStat(node,"Network","SPTM",ANY_DEST,-1,buf);


    sprintf(buf, "Node %u - Number join's sent = %i", sptNode->id, sptNode->nodeStats.joinQueriesSent);
    IO_PrintStat(node,"Network","SPTM",ANY_DEST,-1,buf);


    sprintf(buf, "Node %u - Number leave's sent = %i", sptNode->id, sptNode->nodeStats.leaveQueriesSent);
    IO_PrintStat(node,"Network","SPTM",ANY_DEST,-1,buf);


    sprintf(buf, "Node %u - Number of packets sent = %i", sptNode->id, sptNode->nodeStats.numPacketsSent);
    IO_PrintStat(node,"Network","SPTM",ANY_DEST,-1,buf);


    sprintf(buf, "Node %u - Number of packets received = %i", sptNode->id, sptNode->nodeStats.numPacketsReceived);
    IO_PrintStat(node,"Network","SPTM",ANY_DEST,-1,buf);


    sprintf(buf, "Node %u - Number of packets forwarded = %i", sptNode->id, sptNode->nodeStats.numPacketsForwarded);
    IO_PrintStat(node,"Network","SPTM",ANY_DEST,-1,buf);


    sprintf(buf, "Node %u - Number of packets dropped = %i", sptNode->id, sptNode->nodeStats.numPacketsDropped);
    IO_PrintStat(node,"Network","SPTM",ANY_DEST,-1,buf);


    // Print out all the one hop neighbours of this node
    SptmLinkedListItem* curr = sptNode->adjacentNodes.head.next;    //First adj node
    for (int i=0;i<GetItemCount(&(sptNode->adjacentNodes));i++)
    {
        sprintf(buf, "Node %u - Neighbour: %u", sptNode->id, ((SptmNode*)curr->data)->id);
        IO_PrintStat(node,"Network","SPTM",ANY_DEST,-1,buf);
        curr = curr->next;
    } //END FOR LOOP
    
} /* END PrintStats FUNCTION */








