/*
filename: multicast_sptm.cpp
last modified: April 28, 2008
author: Marvin Pinto
email: marvin@expr.ca
*/


#ifndef _MULTICAST_SPTM_H_
#define _MULTICAST_SPTM_H_



// Debugging Levels
#define SPTM_DEBUG 0
#define SPTM_DEBUG_L2 0
#define SPTM_DEBUG_L3 0

//#define SPTM_DEBUG 1
//#define SPTM_DEBUG_L2 1
//#define SPTM_DEBUG_L3 1



#define SPTM_HELLO_INTERVAL  (1 * SECOND)
#define SPTM_HELLO_DURATION  (1 * SECOND)
#define SPTM_DELAY_JITTER    (1 * MILLI_SECOND)
#define IPOPT_SPTM  124


/* Each item of the node linked list */
typedef struct struct_sptm_linked_list_item
{
	void *data;
    struct struct_sptm_linked_list_item *prev;
    struct struct_sptm_linked_list_item *next;
} SptmLinkedListItem;


/* Data structure for a nodes linked list*/
typedef struct struct_sptm_linked_list
{
    int size;
    SptmLinkedListItem head;
    SptmLinkedListItem tail;
} SptmLinkedList;


/* Sptm simulation statistics data structure */
typedef struct
{
	float avgHopCount;
	int totalNumForwarders;
	int totalSenders;   //actual # of senders in the simulation (not counting stranded nodes)
	int totalReceivers; //actual # of receivers in the simulation (not counting stranded nodes)
	int configuredSenders;  //# of senders in config file
	int configuredReceivers;    //# of receivers in config file
} SptmStats;


/* Node statistics data structure */
typedef struct
{
	int hopCount;              //The hopcout from this node (receiver) to the sender
	char isNodeAForwarder;     //TRUE OR FALSE
	char isNodeASender;        //TRUE OR FALSE
	int numHelloSent;
	int numHelloReceived;
	int joinQueriesSent;
	int leaveQueriesSent;
	int numPacketsReceived;    // Incr everytime the node receives a non duplicate packet
	int numPacketsForwarded;   // Incr everytime the node forwards a non duplicate packet
	int numPacketsSent;        // Num packets a sender has sent (only applies to sender)
	int numPacketsDropped;     // If node is not a forwarder, it drops the packet
	char isNodeAReceiver;       // TRUE OR FALSE //*
} SptmNodeStats;



// Multicast senders add this option to the IP header
// in order to sequence packets.
typedef struct
{
    unsigned long seqNo;
} SptmIpHeaderOption;


// Each node will keep track of the last message its seen
// from for all senders, elimintes duplicate messages.
typedef struct
{
	NodeAddress id;    //multicast sender
    unsigned long seqNo;
} SptmMessageCache;



/* Sptm node data structure */
typedef struct struct_Sptm_Node
{
	NodeAddress id;     // Node ID
	unsigned long weight;      /* Dijkstra */
	char isCovered;      /* Dijkstra */
	struct struct_Sptm_Node *prevNode;		/* Dijkstra */
	SptmLinkedList adjacentNodes;
	SptmNodeStats nodeStats;
	double xcoord;
	double ycoord;
	double zcoord;
	SptmLinkedList msgCache;
	SptmLinkedList forwarderForGroup;   //keep track of which group this node
	                                    //is a forwarder for (if it is a forwarder)
} SptmNode;




/* Hello broadcast message data structure */
typedef struct
{
	SptmNode *sendingNode;
} SptmHello;



// Data structure to store the membership info for a multicast group
typedef struct
{
    NodeAddress id;   //multicast group address
    SptmNode *sender;
    SptmLinkedList receivers;
} SptmMcastGrp;


/* Sptm protocol data structure (state variables) */
typedef struct
{
	char collectStats;   /* TRUE OR FALSE */
	char processHello;      /* When the timer goes off, this value remains FALSE */
	unsigned long maxRange;	/* defaults to 350m */
	int helloDuration;	/* defaults to 75 seconds */
	unsigned long infty;
	int delayJitter;   // defaults to 5 milliseconds times rand value
	SptmStats simulationStats;
	SptmLinkedList sptmNodes;    // Linked list to store all the nodes in the simulation
	SptmLinkedList mcastGroups; // to store all the multicast groups
	unsigned long sequenceNumber;  //to generate a unique msg sequence #
} SptmData;



void SptmInit(Node* node, const NodeInput* nodeInput, int interfaceIndex);
void SptmJoinGroup(Node *node, NodeAddress mcastAddr);
void SptmLeaveGroup(Node *node, NodeAddress mcastAddr);
void SptmRouterFunction(Node *node, Message *msg, NodeAddress destAddr,
    				  int interfaceIndex, BOOL *PacketWasRouted,
    				  NodeAddress prevHop);
void SptmHandleProtocolEvent(Node* node, Message* msg);
void SptmHandleProtocolPacket(Node* node, Message* msg);
void SptmFinalize(Node *node);



static void ReadMulticastSendersFromInputFile(const NodeInput *nodeInput);
static void SptmSetTimer(Node* node, int eventType, clocktype delay);
static void SptmInitializeConfigurableParameters(Node* node, const NodeInput* nodeInput);
static SptmNode* AssociateWithSptmNode(NodeAddress nodeId);
static SptmNode* AssociateWithSptmNode(Node *node);
static void InitializeAllGraphNodes(Node* node, int interfaceIndex);
static void BuildDijkstraTrees(Node* node);
static float MarkForwardingNodes(SptmMcastGrp* group);
static void UpdateNodeWeights(SptmNode *sender);
static SptmLinkedListItem* IsItemInList( void* itemToSearch, 
                                         SptmLinkedList* list, char* type);
static SptmLinkedListItem* IsItemInList( NodeAddress* itemToSearch, 
                                         SptmLinkedList* list, char* type);
static void InitializeList(SptmLinkedList* list);
static SptmLinkedListItem* RemoveFromList( void* itemToRemove, SptmLinkedList* list, char* type);
static char AddToList( void* itemToAdd, SptmLinkedList* list, char* type);
static int GetItemCount( SptmLinkedList* list);
static unsigned long CalculateDistance(SptmNode *n1, SptmNode *n2);
static void PrintStats(Node *node);
static void PrintMulticastGroups(SptmMcastGrp* group);
static void SumSimulationStats();
static unsigned long GetUniqueSequenceNumber();
static void SptmSetIpOption(Message* msg, const SptmIpHeaderOption* opt);
static SptmIpHeaderOption SptmGetIpHeaderOption(const Message* msg);


#endif /* _MULTICAST_SPTM_H_ */

