#include <stdlib.h>
#include <stdio.h>

#include "api.h"
#include "network_ip.h"
#include "routing_srip.h"


void SripInit(Node*            node,
              SripData**       sripPtr,
              const NodeInput* nodeInput,
              int              interfaceIndex)
{
   int i; BOOL retVal;
   Message* newMsg;
   SripData* srip;

   if (MAC_IsWiredNetwork(node, interfaceIndex))
      ERROR_ReportError("SRIP supports only wireless interfaces");

   if (node->numberInterfaces > 1)
      ERROR_ReportError("SRIP only supports one interface of node");

   /* allocate memory for SRIP's variables for this node */
   srip = (SripData*) MEM_malloc(sizeof(SripData));
   (*sripPtr) = srip;

   /* read parameter from the configuration file */
   IO_ReadTime(node->nodeId,
               ANY_ADDRESS,
               nodeInput,
               "SRIP-UPDATE-INTERVAL",
               &retVal,
               &(srip->updateInterval));
   if (retVal == FALSE)
      ERROR_ReportError("SRIP-UPDATE-INTERVAL not specified!");

   /*
    * allocate and initialize the routing table for this node
    *
    * Note: an array of size (numNodes+1) is allocated for convenience since
    * nodeId starts from 1 to numNodes in Qualnet.  The first entry (index 0) is
    * unused.
    */
   srip->routingTable = (SripTableEntry*)
       MEM_malloc(sizeof(SripTableEntry)*(node->numNodes+1));
   for (i = 1; i <= node->numNodes; i++) {
      srip->routingTable[i].destination = i;
      srip->routingTable[i].nextHop     = INVALID_ADDRESS;
      srip->routingTable[i].distance    = SRIP_INFINITY;
   }
   srip->routingTable[node->nodeId].nextHop  = node->nodeId;
   srip->routingTable[node->nodeId].distance = 0;

   /* Initialize statistic */
   srip->numRouteUpdatesBroadcast = 0;

   /* Tell IP to use our function to route packets */
   NetworkIpSetRouterFunction(node, &SripRouterFunction, interfaceIndex);

   /* schedule the very first route update broadcast */
   newMsg = MESSAGE_Alloc(node, NETWORK_LAYER,
                ROUTING_PROTOCOL_SRIP, MSG_NETWORK_RTBroadcastAlarm);
   
   RandomSeed randomSeed;
   RANDOM_SetSeed(randomSeed, node->globalSeed, node->nodeId, ROUTING_PROTOCOL_SRIP);
   MESSAGE_Send(node, newMsg, RANDOM_nrand(randomSeed) % srip->updateInterval);
   //printf("sripinit successful!\n");
}

void SripHandleProtocolEvent(Node* node, Message* msg)
{
   //printf("sriphandleprotocolevent initialize..\n");
   int i, numEntries = 0, pktSize;
   Message*  newMsg;
   char*     pktPtr;

   /* Obtain a pointer to the local variable space */
   SripData* srip = (SripData*)
      NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_SRIP);

   /* Only one message type is used by SRIP */
   assert(msg->eventType == MSG_NETWORK_RTBroadcastAlarm);

   /* count the number of valid entries and calculate the packet size */
   for (i = 1; i <= node->numNodes; i++) {
      if (srip->routingTable[i].distance < SRIP_INFINITY) numEntries++;
   }

   /* prepare a route update packet according to the following format
    *
    *  Packet format
    *  +-------------------+
    *  | Number of entries |
    *  +-------------------+ ---------
    *  |      Dest #1      |     ^
    *  +-------------------+     |
    *  |     nextHop #1    | 1st entry
    *  +-------------------+     |
    *  |    Distance #1    |     v
    *  +-------------------+ ---------
    *  |      Dest #1      |      ^
    *  +-------------------+      |
    *  |     nextHop #1    | 2nd entry
    *  +-------------------+      |
    *  |    Distance #1    |      v
    *  +-------------------+ ---------
    *  |        :          |
    *  +-------------------+ ---------
    *  |      Dest #n      |      ^
    *  +-------------------+      |
    *  |     nextHop #n    | n th entry
    *  +-------------------+      |
    *  |    Distance #n    |      v
    *  +-------------------+ ---------
    *
    */

   newMsg = MESSAGE_Alloc(node, 0, 0, 0);
   pktSize = sizeof(unsigned int) + sizeof(SripTableEntry)*numEntries;
   MESSAGE_PacketAlloc(node, newMsg, pktSize, TRACE_ANY_PROTOCOL);
   pktPtr = newMsg->packet;
   memcpy(pktPtr, &numEntries, sizeof(unsigned int)); /* number of entries */
   pktPtr += sizeof(unsigned int);

   /* Fill the packet with the valid table entries */
   for (i = 1; i <= node->numNodes; i++) {
      if (srip->routingTable[i].distance < SRIP_INFINITY) {
         memcpy(pktPtr, &(srip->routingTable[i]), sizeof(SripTableEntry));
         pktPtr += sizeof(SripTableEntry);
      }
   }

   /* Send the route update packet to MAC layer */
   NetworkIpSendRawMessageToMacLayer(
      node,               /* node pointer */
      newMsg,             /* raw message */
      node->nodeId,       /* source address */
      ANY_DEST,           /* destination address */
      //CONTROL,            /* priority */
      IPTOS_RELIABILITY,
      IPPROTO_SRIP,       /* IP Protocol */
      1,                  /* TTL */
      DEFAULT_INTERFACE,  /* output interface */
      ANY_DEST);          /* next hop address */

   /* update statistic */
   srip->numRouteUpdatesBroadcast++;

   /* schedule the next route update broadcast */
   newMsg = MESSAGE_Alloc(node, NETWORK_LAYER,
                ROUTING_PROTOCOL_SRIP, MSG_NETWORK_RTBroadcastAlarm);
   MESSAGE_Send(node, newMsg, srip->updateInterval);
   //printf("sriphandleprotocolevent successful!..\n");
}

void SripHandleProtocolPacket(Node*       node,
                              Message*    msg,
                              NodeAddress sourceAddress)
{
	//printf("sriphandleprotocolpacket initialize..\n");
   int i, numEntries;
   char *pktPtr;
   SripTableEntry entry;

   /* Obtain a pointer to the local variable space */
   SripData* srip = (SripData*)
      NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_SRIP);

   pktPtr = msg->packet;
   memcpy(&numEntries, pktPtr, sizeof(unsigned int));
   memcpy(pktPtr, &numEntries, sizeof(unsigned int));
   pktPtr += sizeof(unsigned int);

   /* scan the entry list and update routing table only with entries with
    * shorter distance */
   // printf("num entries %d\n", numEntries);
   for (i = 0; i < numEntries; i++)
   {
      memcpy(&entry, pktPtr, sizeof(SripTableEntry));
      entry.distance++;
      if (entry.distance < srip->routingTable[entry.destination].distance) {
         srip->routingTable[entry.destination].distance = entry.distance;
         srip->routingTable[entry.destination].nextHop  = sourceAddress;
      }
      pktPtr += sizeof(SripTableEntry);
   }
	

   /* this is the packet's final destination */
   MESSAGE_Free(node, msg);
   //printf("sriphandleprotocolpacket successful!..\n");
}

void SripFinalize(Node *node)
{
	//printf("sripfinalize initialize..\n");
   char buf[MAX_STRING_LENGTH];

   /* Obtain a pointer to the local variable space */
   SripData* srip = (SripData*)
      NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_SRIP);

   sprintf(buf, "Number of Route Updates Broadcast = %u",
      srip->numRouteUpdatesBroadcast);
   IO_PrintStat(node, "Network", "SRIP", ANY_DEST, -1, buf);
}

void SripRouterFunction(Node*       node,
                        Message*    msg,
                        NodeAddress destAddr,
                        NodeAddress previousHopAddress,
                        BOOL*       packetWasRouted)
{
	//printf("sriprouterfunction initialize..\n");
   IpHeaderType *ipHeader = (IpHeaderType *) msg->packet;

   /* Obtain a pointer to the local variable space */
   SripData* srip = (SripData*)
      NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_SRIP);

   /* do not route any SRIP packet, or any packets destined to myself */
   if (ipHeader->ip_p == IPPROTO_SRIP || ipHeader->ip_dst == node->nodeId)
      return;

   /* route the packet only when the destination is considered reachable */
   if (srip->routingTable[ipHeader->ip_dst].distance < SRIP_INFINITY)
   {
      *packetWasRouted = TRUE;
      NetworkIpSendPacketToMacLayer(
           node,
           msg,
           DEFAULT_INTERFACE,
           srip->routingTable[ipHeader->ip_dst].nextHop);
   }
}
